/*
 * Java
 *
 * Copyright 2021-2022 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.kf;

import com.is2t.elflw.allocator.MemoryBlock;
import com.is2t.elflw.nodes.AllocationSection;
import com.is2t.elflw.nodes.ProgAllocationSection;
import com.is2t.elflw.nodes.RelocationSection;
import com.is2t.elflw.nodes.Section;
import com.is2t.elflw.nodes.SymbolTableSection;

import ej.annotation.Nullable;
import ej.bon.Constants;
import ej.kf.Feature;
import ej.kf.InvalidFormatException;

/**
 * Dynamic load and link of a {@link Feature}. Data is an ELF file generated by SOARv1 in "feature mode", and partially
 * linked with IS2T partial linker. All symbols related to Kernel symbols are already resolved. This dynamic linker has
 * to resolve only its owned symbols and sections.
 */
public class DynamicFeatureLoader extends CommonDynamicFeatureLoader {

	public static final int SUPPORTED_ABI_VERSION = 0x00;

	@Nullable
	private long[] sectionStartAddresses;

	@Override
	protected int getSupportedABIVersion() {
		return SUPPORTED_ABI_VERSION;
	}

	@Override
	protected void resolveRelocations(Section[] sections) throws InvalidFormatException {
		// We compute here once the start of each sections therefore the native relocations will only use theses
		// values if needed
		int nbSections = sections.length;
		long[] sectionStartAddresses = new long[nbSections];
		for (int i = nbSections; --i >= 0;) {
			Section section = sections[i];
			if (section instanceof AllocationSection) {
				sectionStartAddresses[i] = ((AllocationSection) section).getRelocatedAddress();
			}
		}
		this.sectionStartAddresses = sectionStartAddresses;

		super.resolveRelocations(sections);
	}

	@Override
	protected void applyRelocations(RelocationSection rs, ProgAllocationSection relocatedSection)
			throws InvalidFormatException {
		long[] sectionStartAddresses = this.sectionStartAddresses;
		assert sectionStartAddresses != null;
		MemoryBlock data = relocatedSection.data;
		long memoryBlockDescriptor = ((NativeMemoryBlock) data).memoryBlockDescriptor;

		byte[] section = rs.data;
		boolean withAddend = rs.withAddend;
		int entrySize = rs.entrySize();
		SymbolTableSection symbolTableSection = rs.symbolTable;
		assert symbolTableSection != null;
		byte[] symbolTable = symbolTableSection.data;

		int nbEntries = rs.nbEntries();
		int relocationsChunk = Constants.getInt(RELOC_CHUNK_CONSTANT);
		if (Constants.getBoolean(DEBUG_CONSTANT)) {
			int nbChunks = nbEntries / relocationsChunk;
			int remainingEntries = nbEntries % relocationsChunk;
			System.out.println("Relocations chunk " + relocationsChunk + " nb_chunks " + nbChunks + " remaining "
					+ remainingEntries);
		}

		int startIndex = 0;
		while (true) {
			if (startIndex != 0) {
				// give the hand to other threads of same priority, only if there is more than
				// one loop
				// this avoid to trigger a context scheduling for nothing if the chunk size is
				// bigger than the number of relocations entries
				Thread.yield();
			}
			int endIndex = Math.min(startIndex + relocationsChunk, nbEntries);
			if (startIndex >= endIndex) {
				break;
			}

			if (Constants.getBoolean(DEBUG_CONSTANT)) {
				System.out.println("Apply relocations [" + startIndex + "," + endIndex + "[");
			}

			int res = DynamicLoaderNatives.applyRelocations(memoryBlockDescriptor, section, startIndex, endIndex,
					withAddend, entrySize, this.typeHeaderAddress, symbolTable, sectionStartAddresses);
			if (res != 0) {
				if (res == -1) {
					throw new AssertionError();
				} else {
					throw new InvalidFormatException(res);
				}
			}
			startIndex = endIndex;
		}
	}
}
