/*
 * Copyright (c) 2010, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2010-2022 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.is2t.support.net;

import java.io.IOException;
import java.net.BindException;
import java.net.Inet4Address;
import java.net.Inet6Address;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.SocketException;
import java.net.SocketOptions;
import java.net.UnknownHostException;
import java.util.Enumeration;

import com.is2t.errors.NetErrors;
import com.is2t.support.net.natives.ChannelNatives;
import com.is2t.support.net.util.IPAddressConstants;
import com.is2t.support.net.util.NativeIOExceptionHandler;

import ej.error.Message;
import ej.sni.NativeIOException;

/**
 * Native interface to support configuring of channel to run in a non-blocking manner and support scatter/gather io
 * operations.
 */
public abstract class Channel {

	/**
	 * Our reference implementation uses an integer to store the native file descriptor. Implementations without such
	 * support
	 */
	protected final StateImpl nfd;

	public Channel() {
		// XXX consider adding security check here, so only Classpath
		// code may create instances.
		this.nfd = new StateImpl();
	}

	/**
	 * This constructor is used by the POSIX reference implementation; other virtual machines need not support it.
	 *
	 * <strong>Important:</strong> do not call this in library code that is not specific to Classpath's reference
	 * implementation.
	 *
	 * @param native_fd
	 *            The native file descriptor integer.
	 * @throws IOException
	 */
	public Channel(final int native_fd) throws IOException {
		this();
		setNativeFD(native_fd);
	}

	public void setNativeFD(int fd) throws IOException {
		this.nfd.setNativeFD(fd);
	}

	/**
	 * Check if the socket is closed or if the connection is reset by the peer
	 *
	 * @param onlyOnClose
	 *            if true, check only if the socket is closed.
	 * @throws IOException
	 */
	protected void checkSockConnection(boolean onlyOnClose) throws IOException {
		// Nothing to do. VMChannel does not refer to a socket or server socket
		// Override by VMChannelNet
	}

	// Close.
	public static void close(int fd) throws IOException {
		try {
			ChannelNatives.closeNative(fd);
		} catch (NativeIOException e) {
			throw NativeIOExceptionHandler.generateNativeIOException(e);
		}
	}

	public void close() throws IOException {
		this.nfd.close();
	}

	// Inner classes.

	/**
	 * A wrapper for a native file descriptor integer. This tracks the state of an open file descriptor, and ensures
	 * that
	 *
	 * This class need not be fully supported by virtual machines; if a virtual machine does not use integer file
	 * descriptors, or does and wishes to hide that, then the methods of this class may be stubbed out.
	 *
	 * System-specific classes that depend on access to native file descriptor integers SHOULD declare this fact.
	 */
	public final class StateImpl implements State {

		private int native_fd;
		private boolean valid;
		private boolean closed;

		StateImpl() {
			this.native_fd = -1;
			this.valid = false;
			this.closed = false;
		}

		@Override
		public boolean isValid() {
			return this.valid;
		}

		@Override
		public boolean isClosed() {
			synchronized (this) {
				return this.closed;
			}
		}

		@Override
		public int getNativeFD() throws IOException {
			if (!this.valid) {
				throw new IOException(Message.at(NetErrors.getInstance(), NetErrors.J_EBADF));
			}
			return this.native_fd;
		}

		void setNativeFD(final int native_fd) throws IOException {
			if (this.valid) {
				throw new IOException(Message.at(NetErrors.getInstance(), NetErrors.J_EBADF));
			}
			this.native_fd = native_fd;
			this.valid = true;
		}

		@Override
		public synchronized void close() throws IOException {
			if (!this.valid) {
				throw new IOException(Message.at(NetErrors.getInstance(), NetErrors.J_EBADF));
			}

			if (isStream()) {
				try {
					// Before close socket, call shutdown to unblock the channel
					ChannelNatives.shutdown(this.native_fd);
				} catch (NativeIOException e) {
					throw NativeIOExceptionHandler.generateNativeIOException(e);
				}
			}
			Channel.close(this.native_fd);
			this.valid = false;
			this.closed = true;
		}

		@Override
		public String toString() {
			if (this.closed) {
				return "<<closed>>";
			}
			if (!this.valid) {
				return "<<invalid>>";
			}
			return String.valueOf(this.native_fd);
		}

	}

	/**
	 * Check if this channel refers to a stream channel (Socket/ServerSocket)
	 *
	 * @return
	 */
	public abstract boolean isStream();

	static interface State {
		boolean isValid();

		boolean isClosed();

		int getNativeFD() throws IOException;

		void close() throws IOException;
	}

	public State getState() {
		return this.nfd;
	}

	public void setOption(int optionId, Object optionValue) throws SocketException {
		try {
			int fd = this.nfd.getNativeFD();
			int value;

			if (optionValue instanceof Integer) {
				value = ((Integer) optionValue).intValue();
			} else if (optionValue instanceof Boolean) {
				value = ((Boolean) optionValue).booleanValue() ? 1 : 0;
				// special handling for linger: -1 is a non-error value meaning "disabled"
				if ((value == 0) && (optionId == SocketOptions.SO_LINGER)) {
					value = -1;
				}
			} else if (optionValue instanceof java.net.NetworkInterface) {
				Enumeration<InetAddress> addrs = ((java.net.NetworkInterface) optionValue).getInetAddresses();
				if (!addrs.hasMoreElements()) {
					// Can't set interface: No IP addresses bound to interface
					throw new SocketException(Message.at(NetErrors.getInstance(), NetErrors.J_EINVAL));
				}
				do {
					setOption(optionId, addrs.nextElement());
				} while (addrs.hasMoreElements());

				// need to break from here, as the following block is for the int/boolean values
				return;
			} else if (optionValue instanceof InetAddress) {
				if (!(optionValue instanceof Inet4Address) && !(optionValue instanceof Inet6Address)) {
					throw new SocketException(Message.at(NetErrors.getInstance(), NetErrors.J_EAFNOSUPPORT));
				}
				byte[] addr = ((InetAddress) optionValue).getAddress();
				ChannelNatives.setOption(fd, optionId, addr, addr.length);

				// need to break from here, as the following block is for the int/boolean values
				return;
			} else {
				throw new SocketException(Message.at(NetErrors.getInstance(), NetErrors.J_EINVAL));
			}

			// if we reach here, then we have an int value
			ChannelNatives.setOption(fd, optionId, value);
		} catch (NativeIOException e) {
			throw NativeIOExceptionHandler.generateNativeSocketException(e);
		} catch (IOException e) {
			SocketException ex = new SocketException(e.getMessage());
			ex.initCause(e);
			throw ex;
		}
	}

	/**
	 * Get a socket option. This implementation is only required to support socket options that are boolean values,
	 * which include:
	 *
	 * SocketOptions.IP_MULTICAST_LOOP SocketOptions.SO_BROADCAST SocketOptions.SO_KEEPALIVE SocketOptions.SO_OOBINLINE
	 * SocketOptions.SO_REUSEADDR SocketOptions.TCP_NODELAY
	 *
	 * and socket options that are integer values, which include:
	 *
	 * SocketOptions.IP_TOS SocketOptions.SO_LINGER SocketOptions.SO_RCVBUF SocketOptions.SO_SNDBUF
	 * SocketOptions.SO_TIMEOUT
	 *
	 * @param optionId
	 *            The option ID to fetch.
	 * @return A {@link Boolean} or {@link Integer} containing the socket option.
	 * @throws SocketException
	 */
	public Object getOption(int optionId) throws SocketException {
		switch (optionId) {
		case SocketOptions.IP_MULTICAST_IF:
		case SocketOptions.IP_MULTICAST_IF2:
			return getOptionValueByteArray(optionId);
		default:
			return getOptionValueInt(optionId);
		}
	}

	/**
	 * @param optionId
	 * @return
	 */
	private Object getOptionValueInt(int optionId) throws SocketException {
		int value;
		try {
			value = ChannelNatives.getOption(this.nfd.getNativeFD(), optionId);
		} catch (NativeIOException cause) {
			throw NativeIOExceptionHandler.generateNativeSocketException(cause);
		} catch (IOException e) {
			SocketException ex = new SocketException(e.getMessage());
			ex.initCause(e);
			throw ex;
		}

		if (value < 0) {
			// special handling for linger: -1 is a non-error value meaning "disabled"
			if ((optionId != SocketOptions.SO_LINGER) || (value != -1)) {
				throw new SocketException(Message.at(NetErrors.getInstance(), value));
			}
		}

		switch (optionId) {
		case SocketOptions.IP_MULTICAST_LOOP:
		case SocketOptions.SO_REUSEADDR:
		case SocketOptions.SO_OOBINLINE:
		case SocketOptions.SO_BROADCAST:
		case SocketOptions.SO_KEEPALIVE:
		case SocketOptions.TCP_NODELAY:
			return Boolean.valueOf(value != 0);

		case SocketOptions.IP_TOS:
		case SocketOptions.SO_RCVBUF:
		case SocketOptions.SO_SNDBUF:
		case SocketOptions.SO_TIMEOUT:
		case SocketOptions.SO_LINGER:
			return new Integer(value);

		default:
			throw new SocketException(Message.at(NetErrors.getInstance(), NetErrors.J_ENOPROTOOPT));
		}
	}

	private Object getOptionValueByteArray(int optionId) throws SocketException {
		byte[] optionBuffer = new byte[IPAddressConstants.IPV6_ADDRESS_SIZE];
		try {
			int ret = ChannelNatives.getOption(this.nfd.getNativeFD(), optionId, optionBuffer, optionBuffer.length);
			byte[] addr = new byte[ret];
			System.arraycopy(optionBuffer, 0, addr, 0, addr.length);
			InetAddress inetAddr = InetAddress.getByAddress(addr);
			switch (optionId) {
			case SocketOptions.IP_MULTICAST_IF:
				return inetAddr;
			case SocketOptions.IP_MULTICAST_IF2:
				return java.net.NetworkInterface.getByInetAddress(inetAddr);
			default:
				throw new SocketException(Message.at(NetErrors.getInstance(), NetErrors.J_ENOPROTOOPT));
			}
		} catch (NativeIOException cause) {
			throw NativeIOExceptionHandler.generateNativeSocketException(cause);
		} catch (SocketException cause) {
			throw cause;
		} catch (UnknownHostException cause) {
			SocketException ex = new SocketException(cause.getMessage());
			ex.initCause(cause);
			throw ex;
		} catch (IOException cause) {
			SocketException ex = new SocketException(cause.getMessage());
			ex.initCause(cause);
			throw ex;
		}
	}

	/**
	 * Binds this socket to the given local address and port.
	 *
	 * @param address
	 *            The address to bind to; the InetAddress is either an IPv4 or IPv6 address.
	 * @throws IOException
	 *             If binding fails; for example, if the port in the given InetSocketAddress is privileged, and the
	 *             current process has insufficient privileges.
	 */
	public void bind(InetSocketAddress address) throws IOException {
		InetAddress addr = address.getAddress();
		byte[] addressAsBytes = addr.getAddress();
		try {
			int addrLength = addr.getAddress().length;
			if ((addrLength == IPAddressConstants.IPV4_ADDRESS_SIZE)
					|| (addrLength == IPAddressConstants.IPV6_ADDRESS_SIZE)) {
				ChannelNatives.bind(this.nfd.getNativeFD(), addressAsBytes, addressAsBytes.length, address.getPort());
			} else {
				throw new SocketException(Message.at(NetErrors.getInstance(), NetErrors.J_EAFNOSUPPORT));
			}
		} catch (NativeIOException cause) {
			int errCode = cause.getErrorCode();
			if ((errCode == NetErrors.J_EADDRINUSE) || (errCode == NetErrors.J_EACCES)) {
				BindException ex = new BindException(Message.at(NetErrors.getInstance(), errCode));
				ex.initCause(cause);
				throw ex;
			} else {
				throw NativeIOExceptionHandler.generateNativeSocketException(cause);
			}
		}
	}

	/**
	 * Listen on this socket for incoming connections.
	 *
	 * @param backlog
	 *            The backlog of connections.
	 * @throws IOException
	 *             If listening fails.
	 */
	public void listen(int backlog) throws IOException {
		try {
			ChannelNatives.listen(this.nfd.getNativeFD(), backlog);
		} catch (NativeIOException cause) {
			throw NativeIOExceptionHandler.generateNativeSocketException(cause);
		}
	}

}