/*
 * Copyright (c) 2010, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2010-2022 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.is2t.support.net;

import java.net.UnknownHostException;

import com.is2t.support.net.natives.DNSNatives;
import com.is2t.support.net.natives.NetworkAddressNatives;
import com.is2t.support.net.util.IPAddressConstants;
import com.is2t.support.net.util.NativeIOExceptionHandler;

import ej.sni.NativeIOException;
import ej.sni.SNI;

public class NetworkAddress {

	private static final int HOST_NAME_MAX_SIZE = 253;
	private static final int IP_ADDR_MAX_SIZE = 16; // use ipv6 size as max IP address size

	public static String getLocalHostname() throws UnknownHostException {
		byte[] localHostNameBuffer = new byte[HOST_NAME_MAX_SIZE];
		try {
			NetworkAddressNatives.getLocalHostnameNative(localHostNameBuffer, localHostNameBuffer.length);
			return SNI.toJavaString(localHostNameBuffer);
		} catch (NativeIOException e) {
			throw NativeIOExceptionHandler.generateNativeUnknownHostException(e);
		}
	}

	public static byte[] lookupInaddrAny() throws UnknownHostException {
		byte[] addrBuffer = new byte[IPAddressConstants.IPV6_ADDRESS_SIZE];
		try {
			int addrSize = NetworkAddressNatives.lookupInaddrAny(addrBuffer, addrBuffer.length);
			byte[] lookupInaddrAny = new byte[addrSize];
			System.arraycopy(addrBuffer, 0, lookupInaddrAny, 0, addrSize);
			return lookupInaddrAny;
		} catch (NativeIOException e) {
			throw NativeIOExceptionHandler.generateNativeUnknownHostException(e);
		}
	}

	public static byte[] loopbackAddress() throws UnknownHostException {
		byte[] addrBuffer = new byte[IPAddressConstants.IPV6_ADDRESS_SIZE];
		try {
			int addrSize = NetworkAddressNatives.loopbackAddress(addrBuffer, addrBuffer.length);
			byte[] loopbackAddress = new byte[addrSize];
			System.arraycopy(addrBuffer, 0, loopbackAddress, 0, addrSize);
			return loopbackAddress;
		} catch (NativeIOException e) {
			throw NativeIOExceptionHandler.generateNativeUnknownHostException(e);
		}
	}

	public static byte[][] getHostByName(String hostname) throws UnknownHostException {
		int hostnameLength = hostname.getBytes().length;
		byte[] hostnameBuffer = SNI.toCString(hostname);
		try {
			int nb = DNSNatives.getHostByNameCount(hostnameBuffer, hostnameLength);
			byte[][] res = new byte[nb][];
			byte[] addressBuffer = new byte[IP_ADDR_MAX_SIZE];
			while (--nb >= 0) {
				int hostAddrLen = DNSNatives.getHostByNameAt(nb, hostnameBuffer, hostnameLength, addressBuffer,
						addressBuffer.length);
				System.arraycopy(addressBuffer, 0, res[nb] = new byte[hostAddrLen], 0, hostAddrLen);
			}
			return res;
		} catch (NativeIOException e) {
			throw NativeIOExceptionHandler.generateNativeUnknownHostException(e);
		}
	}

	public static String getHostByAddr(byte[] ip) throws UnknownHostException {
		try {
			byte[] hostname = new byte[HOST_NAME_MAX_SIZE];
			DNSNatives.getHostByAddr(ip, ip.length, hostname, hostname.length);
			return SNI.toJavaString(hostname);
		} catch (NativeIOException e) {
			throw NativeIOExceptionHandler.generateNativeUnknownHostException(e);
		}
	}

}
