/*
 * Copyright (c) 2010, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2010-2022 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.is2t.support.net;

import java.net.SocketException;

import com.is2t.errors.NetErrors;
import com.is2t.support.net.natives.NetworkInterfaceNatives;
import com.is2t.support.net.util.IPAddressConstants;
import com.is2t.support.net.util.NativeIOExceptionHandler;

import ej.error.Message;
import ej.sni.NativeIOException;
import ej.sni.SNI;

public final class NetworkInterfaceSupport {

	// ipv4 address info size (tag (1) + IP (4) + prefix (1) + hasBroadcast (1) +
	// broadcast IP (4))
	public static final int IPV4_ADDR_INFO_SIZE = 11;
	// ipv6 address info size (tag (1) + IP (16) + prefix (1)
	public static final int IPV6_ADDR_INFO_SIZE = 18;

	// ipv4 address tag
	public static final int IPV4_ADDR_TAG = 4;
	// ipv6 address tag
	public static final int IPV6_ADDR_TAG = 6;

	// Hardware address size
	private static final int MAC_ADDR_SIZE = 6;

	private static final int INTERFACE_MAX_NAME_LENGTH = 128;

	/**
	 * Return a list of VM network interface objects.
	 *
	 * @return The list of network interfaces.
	 * @throws SocketException
	 */
	public static VMNetworkInterfaceData[] getVMInterfaces() throws SocketException {
		try {
			int nIfaces = NetworkInterfaceNatives.getVMInterfacesCount();
			VMNetworkInterfaceData[] res = new VMNetworkInterfaceData[nIfaces];
			byte[] ifnameBuffer = new byte[INTERFACE_MAX_NAME_LENGTH];
			byte[] addrInfo = new byte[IPV6_ADDR_INFO_SIZE];
			for (int i = nIfaces; --i >= 0;) {
				int ifnameLen = NetworkInterfaceNatives.getVMInterface(i, ifnameBuffer, ifnameBuffer.length);
				if (ifnameLen <= 0) {
					// no interface found
					continue;
				}
				String ifname = SNI.toJavaString(ifnameBuffer);
				ifnameLen = ifname.getBytes().length;
				int count = NetworkInterfaceNatives.getVMInterfaceAddressesCount(i, ifnameBuffer, ifnameLen);
				byte[][] addresses = new byte[count][];
				short[] prefixes = new short[count];
				byte[][] broadcasts = new byte[count][];

				while (--count >= 0) {
					int size = NetworkInterfaceNatives.getVMInterfaceAddress(i, ifnameBuffer, ifnameLen, count,
							addrInfo, addrInfo.length);
					int addrTag;
					int addrSize;
					boolean addrMayHaveBroadcast;
					if (size == IPV4_ADDR_INFO_SIZE) {
						addrTag = IPV4_ADDR_TAG;
						addrSize = IPAddressConstants.IPV4_ADDRESS_SIZE;
						addrMayHaveBroadcast = true;
					} else if (size == IPV6_ADDR_INFO_SIZE) {
						addrTag = IPV6_ADDR_TAG;
						addrSize = IPAddressConstants.IPV6_ADDRESS_SIZE;
						addrMayHaveBroadcast = false;
					} else {
						throw new SocketException(Message.at(NetErrors.getInstance(), NetErrors.J_EAFNOSUPPORT));
					}
					if (addrInfo[0] != addrTag) {
						throw new SocketException(Message.at(NetErrors.getInstance(), NetErrors.J_EAFNOSUPPORT));
					}
					byte[] address = new byte[addrSize];
					// +1 because of the tag
					System.arraycopy(addrInfo, 1, address, 0, addrSize);
					addresses[count] = address;
					// prefix can be from 0 to 128
					// so we encode it on 1 byte, from 0 to 0x80
					prefixes[count] = (short) (addrInfo[1 + addrSize] & 0xFF);

					byte[] broadcast;
					if (addrMayHaveBroadcast) {
						if (addrInfo[addrSize + 2] != 0) {
							broadcast = new byte[addrSize];
							// +1 because of the tag
							// +1 because of the prefix
							// +1 because of hasBroadcast
							System.arraycopy(addrInfo, addrSize + 3, broadcast, 0, addrSize);
						} else {
							broadcast = null;
						}
					} else {
						broadcast = null;
					}
					broadcasts[count] = broadcast;
				}
				res[i] = new VMNetworkInterfaceData(i, ifname, addresses, prefixes, broadcasts);
			}
			return res;
		} catch (NativeIOException e) {
			throw handleNativeException(e);
		}
	}

	public static boolean isUp(String name) throws SocketException {
		int nameLength = name.getBytes().length;
		byte[] nameBuffer = SNI.toCString(name);
		try {
			int ret = NetworkInterfaceNatives.isUp(nameBuffer, nameLength);
			return (ret == 0); // 0 means true
		} catch (NativeIOException e) {
			throw handleNativeException(e);
		}
	}

	public static boolean isLoopback(String name) throws SocketException {
		int nameLength = name.getBytes().length;
		byte[] nameBuffer = SNI.toCString(name);
		try {
			int ret = NetworkInterfaceNatives.isLoopback(nameBuffer, nameLength);
			return (ret == 0); // 0 means true
		} catch (NativeIOException e) {
			throw handleNativeException(e);
		}
	}

	public static boolean isPointToPoint(String name) throws SocketException {
		int nameLength = name.getBytes().length;
		byte[] nameBuffer = SNI.toCString(name);
		try {
			int ret = NetworkInterfaceNatives.isPointToPoint(nameBuffer, nameLength);
			return (ret == 0); // 0 means true
		} catch (NativeIOException e) {
			throw handleNativeException(e);
		}
	}

	public static boolean supportsMulticast(String name) throws SocketException {
		int nameLength = name.getBytes().length;
		byte[] nameBuffer = SNI.toCString(name);
		try {
			int ret = NetworkInterfaceNatives.supportsMulticast(nameBuffer, nameLength);
			return (ret == 0); // 0 means true
		} catch (NativeIOException e) {
			throw handleNativeException(e);
		}
	}

	public static byte[] getHardwareAddress(String name) throws SocketException {
		int nameLength = name.getBytes().length;
		byte[] nameBuffer = SNI.toCString(name);
		byte[] hwAddr = new byte[MAC_ADDR_SIZE];
		try {
			int ret = NetworkInterfaceNatives.getHardwareAddress(nameBuffer, nameLength, hwAddr, hwAddr.length);
			if (ret == 0) {
				return null;
			} else if (ret == MAC_ADDR_SIZE) {
				return hwAddr;
			} else {
				throw new SocketException(Message.at(NetErrors.getInstance(), NetErrors.J_WRONG_MAC_ADDRESS_SIZE));
			}
		} catch (NativeIOException e) {
			throw handleNativeException(e);
		}
	}

	public static int getMTU(String name) throws SocketException {
		int nameLength = name.getBytes().length;
		byte[] nameBuffer = SNI.toCString(name);
		try {
			return NetworkInterfaceNatives.getMTU(nameBuffer, nameLength);
		} catch (NativeIOException e) {
			throw handleNativeException(e);
		}
	}

	private static SocketException handleNativeException(NativeIOException cause) {
		int errCode = cause.getErrorCode();
		if (errCode == NetErrors.J_NET_NATIVE_NOT_IMPLEMENTED) {
			throw new RuntimeException(Message.at(NetErrors.getInstance(), errCode));
		} else {
			return NativeIOExceptionHandler.generateNativeSocketException(cause);
		}
	}
}
