/*
 * Copyright (c) 2010, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2010-2022 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.is2t.support.net;

import java.io.IOException;
import java.net.ConnectException;
import java.net.Inet4Address;
import java.net.Inet6Address;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.NoRouteToHostException;
import java.net.SocketException;
import java.net.SocketOptions;
import java.net.SocketTimeoutException;

import com.is2t.errors.NetErrors;
import com.is2t.support.net.natives.SocketChannelNatives;
import com.is2t.support.net.util.IPAddressConstants;
import com.is2t.support.net.util.NativeIOExceptionHandler;

import ej.bon.Util;
import ej.error.Message;
import ej.sni.NativeIOException;

public abstract class SocketChannel extends Channel {

	public SocketChannel() throws IOException {
		super();
	}

	public SocketChannel(int fd) throws IOException {
		super(fd);
	}

	// Network (socket) specific methods.

	/**
	 * Connect the underlying socket file descriptor to the remote host.
	 *
	 * @param saddr
	 *            The address to connect to.
	 * @param timeout
	 *            The connect timeout to use for blocking connects.
	 * @throws SocketException
	 *             If an error occurs while connecting.
	 * @throws SocketTimeoutException
	 *             If timeout expires before connecting.
	 */
	public void connect(InetSocketAddress saddr, int timeout) throws SocketException, SocketTimeoutException {
		int fd;
		long absoluteTimeout = 0;
		InetAddress addr = saddr.getAddress();

		try {
			fd = this.nfd.getNativeFD();
		} catch (IOException e) {
			SocketException ex = new SocketException(e.getMessage());
			ex.initCause(e);
			throw ex;
		}

		try {
			byte[] hostAddrBuffer = addr.getAddress();
			int addrLength = addr.getAddress().length;
			if ((addrLength == IPAddressConstants.IPV4_ADDRESS_SIZE)
					|| (addrLength == IPAddressConstants.IPV6_ADDRESS_SIZE)) {
				if (timeout > 0) {
					absoluteTimeout = Util.platformTimeMillis() + timeout;
				}
				SocketChannelNatives.connect(fd, hostAddrBuffer, hostAddrBuffer.length, saddr.getPort(),
						absoluteTimeout);
			} else {
				throw new SocketException(Message.at(NetErrors.getInstance(), NetErrors.J_EAFNOSUPPORT));
			}
		} catch (NativeIOException cause) {
			int errCode = cause.getErrorCode();
			// Check if timeout reached
			if (errCode == NetErrors.J_ETIMEDOUT) {
				SocketTimeoutException ex = new SocketTimeoutException(
						Message.at(NetErrors.getInstance(), NetErrors.J_ETIMEDOUT));
				ex.initCause(cause);
				throw ex;
			}
			// Convert to other Exception
			if (errCode == NetErrors.J_ECONNREFUSED) {
				ConnectException ex = new ConnectException(Message.at(NetErrors.getInstance(), errCode));
				ex.initCause(cause);
				throw ex;
			} else if (errCode == NetErrors.J_EHOSTUNREACH) {
				NoRouteToHostException ex = new NoRouteToHostException(Message.at(NetErrors.getInstance(), errCode));
				ex.initCause(cause);
				throw ex;
			} else if (errCode == NetErrors.J_NET_NATIVE_NOT_IMPLEMENTED) {
				throw new RuntimeException(Message.at(NetErrors.getInstance(), errCode));
			} else {
				throw NativeIOExceptionHandler.generateNativeSocketException(cause);
			}
		}
	}

	public InetSocketAddress getLocalAddress() throws IOException {
		if (!this.nfd.isValid()) {
			return null;
		}

		int namelen, port;
		byte[] localAddrBuffer = new byte[IPAddressConstants.IPV6_ADDRESS_SIZE];
		int native_fd = this.nfd.getNativeFD();
		try {
			namelen = SocketChannelNatives.getLocalAddress(native_fd, localAddrBuffer, localAddrBuffer.length);
			if (namelen == 0) {
				throw new SocketException(Message.at(NetErrors.getInstance(), NetErrors.J_SOCKET_NOT_BOUND));
			}
			port = SocketChannelNatives.getLocalPort(native_fd);
			if (namelen == IPAddressConstants.IPV4_ADDRESS_SIZE) {
				byte[] tmpBuff = new byte[IPAddressConstants.IPV4_ADDRESS_SIZE];
				System.arraycopy(localAddrBuffer, 0, tmpBuff, 0, IPAddressConstants.IPV4_ADDRESS_SIZE);
				return new InetSocketAddress(Inet4Address.getByAddress(tmpBuff), port);
			} else if (namelen == IPAddressConstants.IPV6_ADDRESS_SIZE) {
				byte[] tmpBuff = new byte[IPAddressConstants.IPV6_ADDRESS_SIZE];
				System.arraycopy(localAddrBuffer, 0, tmpBuff, 0, IPAddressConstants.IPV6_ADDRESS_SIZE);
				return new InetSocketAddress(Inet6Address.getByAddress(tmpBuff), port);
			}
			throw new SocketException(Message.at(NetErrors.getInstance(), NetErrors.J_EAFNOSUPPORT));
		} catch (NativeIOException cause) {
			int errCode = cause.getErrorCode();
			if (errCode == -1) {
				SocketException ex = new SocketException(
						Message.at(NetErrors.getInstance(), NetErrors.J_SOCKET_NOT_BOUND));
				ex.initCause(cause);
				throw ex;
			} else {
				throw NativeIOExceptionHandler.generateNativeSocketException(cause);
			}
		}
	}

	/**
	 * Returns the socket address of the remote peer this channel is connected to
	 *
	 * @return The peer address.
	 * @throws IOException
	 */
	public InetSocketAddress getPeerAddress() throws IOException {
		if (!this.nfd.isValid()) {
			return null;
		}
		int namelen;
		int native_fd = this.nfd.getNativeFD();
		byte[] peerAddr = new byte[IPAddressConstants.IPV6_ADDRESS_SIZE];
		try {
			namelen = SocketChannelNatives.getPeerAddress(native_fd, peerAddr, peerAddr.length);
			if (namelen == 0) {
				throw new SocketException(Message.at(NetErrors.getInstance(), NetErrors.J_ENOTCONN));
			}
			int port = SocketChannelNatives.getPeerPort(native_fd);
			if (namelen == IPAddressConstants.IPV4_ADDRESS_SIZE) {
				byte[] tmpBuff = new byte[IPAddressConstants.IPV4_ADDRESS_SIZE];
				System.arraycopy(peerAddr, 0, tmpBuff, 0, IPAddressConstants.IPV4_ADDRESS_SIZE);
				return new InetSocketAddress(Inet4Address.getByAddress(tmpBuff), port);
			} else if (namelen == IPAddressConstants.IPV6_ADDRESS_SIZE) {
				byte[] tmpBuff = new byte[IPAddressConstants.IPV6_ADDRESS_SIZE];
				System.arraycopy(peerAddr, 0, tmpBuff, 0, IPAddressConstants.IPV6_ADDRESS_SIZE);
				return new InetSocketAddress(Inet6Address.getByAddress(tmpBuff), port);
			}
			throw new SocketException(Message.at(NetErrors.getInstance(), NetErrors.J_EAFNOSUPPORT));
		} catch (NativeIOException cause) {
			throw NativeIOExceptionHandler.generateNativeSocketException(cause);
		}
	}

	@Override
	protected void checkSockConnection(boolean onlyOnClose) throws SocketException {
		if (this.nfd.isClosed()) {
			// closed socket
			throw new SocketException(Message.at(NetErrors.getInstance(), NetErrors.J_SOCKET_CLOSED));
		}
		if (!onlyOnClose) {
			// check if the channel is still connected
			InetSocketAddress peer = null;
			try {
				peer = getPeerAddress();
			} catch (IOException e) {

			} finally {
				if (peer == null) {
					// connection reset
					throw new SocketException(Message.at(NetErrors.getInstance(), NetErrors.J_ECONNRESET));
				}
			}
		}
	}

	/**
	 * Create a new socket. This method will initialize the native file descriptor state of this instance.
	 *
	 * @param stream
	 *            Whether or not to create a streaming socket, or a datagram socket.
	 * @throws IOException
	 *             If creating a new socket fails, or if this channel already has its native descriptor initialized.
	 */
	public void initSocket(boolean stream) throws IOException {
		if (this.nfd.isValid()) {
			throw new IOException(Message.at(NetErrors.getInstance(), NetErrors.J_NATIVE_FD_ALREADY_INITIALIZE));
		}

		try {
			int fd = SocketChannelNatives.socket(stream);
			setNativeFD(fd);
		} catch (NativeIOException cause) {
			throw new IOException(Message.at(NetErrors.getInstance(), cause.getErrorCode()), cause);
		}
	}

	public void initServerSocket() throws java.io.IOException {
		if (this.nfd.isValid()) {
			throw new IOException(Message.at(NetErrors.getInstance(), NetErrors.J_NATIVE_FD_ALREADY_INITIALIZE));
		}

		try {
			int fd = SocketChannelNatives.serverSocket();
			setNativeFD(fd);
		} catch (NativeIOException cause) {
			throw NativeIOExceptionHandler.generateNativeSocketException(cause);
		}
	}

	public void initMulticastSocket() throws IOException {
		if (this.nfd.isValid()) {
			throw new IOException(Message.at(NetErrors.getInstance(), NetErrors.J_NATIVE_FD_ALREADY_INITIALIZE));
		}

		try {
			int fd = SocketChannelNatives.multicastSocket();
			setNativeFD(fd);
		} catch (NativeIOException cause) {
			throw NativeIOExceptionHandler.generateNativeSocketException(cause);
		}
	}

	@Override
	public Object getOption(int optionId) throws SocketException {
		if (optionId == SocketOptions.SO_BINDADDR) {
			try {
				InetSocketAddress local = getLocalAddress();
				if (local == null) {
					return null;
				}
				return local.getAddress();
			} catch (SocketException se) {
				throw se;
			} catch (IOException ioe) {
				SocketException se = new SocketException();
				se.initCause(ioe);
				throw se;
			}
		}
		return super.getOption(optionId);
	}
}
