/*
 * Copyright (c) 2014, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2014-2022 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.is2t.support.net.natives;

import com.is2t.errors.NetErrors;

import ej.sni.NativeIOException;

public class DatagramSocketChannelNatives {

	/**
	 * Receives data from the datagram socket associated with the file descriptor {@code fd}.
	 * <p>
	 * When this method returns, the {@code remoteHost} buffer contains the remote host address followed by the remote
	 * host port from which the data is received. Remote host port is an int.
	 *
	 * The receive operation will block until data is arrived, the given absolute timeout is reached, or an error
	 * occurs. If the timeout expires, an <B>ej.sni.NativeIOException</B> is raised with the error code
	 * <B>NetErrors.J_ETIMEDOUT</B>.
	 * 
	 * {@code absoluteTimeout} is an absolute time in milliseconds computed from the system time returned by
	 * <B>ej.bon.Util.platformTimeMillis()</B>. A timeout of zero is interpreted as an infinite timeout.
	 * 
	 * @param fd
	 *            the datagram socket file descriptor
	 * @param dst
	 *            the destination buffer into which the received data will be stored
	 * @param dstOffset
	 *            the offset in the destination buffer
	 * @param dstLength
	 *            the maximum number of data to receive
	 * @param remoteHost
	 *            the buffer into which the remote host address and port from which the data is received will be stored
	 * @param remoteHostLength
	 *            the maximum number of data in the remote host buffer
	 * @param absoluteTimeout
	 *            the absolute timeout in milliseconds or 0 if no timeout.
	 * @return long value on success, where the 32 most significant bits converted in integer is the number of data
	 *         received (in bytes) and the 32 least significant bits is the remote host address size (in bytes); or -1
	 *         when end-of-file is encountered.
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native long receive(int fd, byte[] dst, int dstOffset, int dstLength, byte[] remoteHost,
			int remoteHostLength, long absoluteTimeout) throws NativeIOException;

	/**
	 * Sends data from the datagram socket associated with the file descriptor {@code fd} to the remote host address
	 * {@code dstAddr} and {@code port}. The send operation will block until all data have been sent or an error occurs.
	 * 
	 * @param fd
	 *            the datagram socket file descriptor
	 * @param data
	 *            the buffer of data to send
	 * @param dataOffset
	 *            the offset in the data buffer
	 * @param dataLength
	 *            the number of data to send
	 * @param dstAddr
	 *            the remote host IP address at which the data will be sent
	 * @param dstAddrLength
	 *            the remote host address size
	 * @param dstPort
	 *            the remote host port
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native void send(int fd, byte[] data, int dataOffset, int dataLength, byte[] dstAddr,
			int dstAddrLength, int dstPort) throws NativeIOException;

	/**
	 * Disconnects the datagram socket associated with the file descriptor {@code fd}.
	 *
	 * @param fd
	 *            the datagram socket file descriptor
	 * @throws NativeIOException
	 *             if an error occurs.
	 */
	public static native void disconnect(int fd) throws NativeIOException;

}
