/*
 * Copyright (c) 2014, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2014-2022 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.is2t.support.net.natives;

import com.is2t.errors.NetErrors;
import com.is2t.support.net.NetworkInterfaceSupport;

import ej.sni.NativeIOException;

public class NetworkInterfaceNatives {

	static {
		ChannelNatives.initializeNativePart();
	}

	/**
	 * Gets the hardware address associated with the given network interface name {@code ifname}.
	 * <p>
	 * When this method returns, the buffer {@code hwAddr} is filled with the hardware address.
	 *
	 * @param ifname
	 *            the network interface name (null-terminated string).
	 * @param ifnameLength
	 *            the network interface name length. (The length does not include the terminating null byte).
	 * @param hwAddr
	 *            the buffer into which the hardware address will be stored
	 * @param hwAddrMaxLength
	 *            maximum size of hardware address that can be stored in the buffer {@code hwAddr}
	 * @return the hardware address size in bytes, 0 if no hardware address found.
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native int getHardwareAddress(byte[] ifname, int ifnameLength, byte[] hwAddr, int hwAddrMaxLength)
			throws NativeIOException;

	/**
	 * Gets the network interface name at the {@code index} position in the network interfaces list. When this method
	 * returns, the {@link ifnameBuffer} buffer is filled with the network interface name
	 *
	 * @param index
	 *            the index-position of the interface in the interfaces list.
	 * @param ifnameBuffer
	 *            the buffer into which the network interface name will be stored. (The stored name must be
	 *            null-terminated string).
	 * @param ifnameBufferLength
	 *            the length of the network interface buffer
	 * @return the network interface name length (in bytes); 0 or negative value if no interface found. (The returned
	 *         length must not include the terminating null byte).
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native int getVMInterface(int index, byte[] ifnameBuffer, int ifnameBufferLength)
			throws NativeIOException;

	/**
	 * Returns the number of the network interface addresses.
	 * 
	 * @param index
	 *            the index-position of the interface in the interfaces list.
	 * @param ifname
	 *            the network interface name (null-terminated string).
	 * @param ifnameLength
	 *            the network interface name length. (The length does not include the terminating null byte).
	 * @return the number of the network interface addresses.
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native int getVMInterfaceAddressesCount(int index, byte[] ifname, int ifnameLength)
			throws NativeIOException;

	/**
	 * Gets a network interface address in the interface addresses list.
	 * 
	 * <p>
	 * When this method returns, the {@code addrInfo} buffer is filled with the network interface address information,
	 * using the following layout:
	 * <ul>
	 * <li>a one-byte tag {@link NetworkInterfaceSupport#IPV4_ADDR_TAG IPV4_ADDR_TAG} or
	 * {@link NetworkInterfaceSupport#IPV6_ADDR_TAG IPV6_ADDR_TAG}</li>
	 * <li>the IP address bytes (4 or 16)</li>
	 * <li>the prefix length in bits (encoded values from 0 to 0x80, meaning 1 to 128)</li>
	 * <li>for an IP v4 address, a one-byte boolean value (0 or 1) telling if there's a broadcast address</li>
	 * <li>for an IP v4 address that has broadcast, the 4 bytes of the broadcast address</li>
	 * </ul>
	 *
	 * @param idxIf
	 *            the index-position of the interface in the interfaces list.
	 * @param ifname
	 *            the network interface name (null-terminated string).
	 * @param ifnameLength
	 *            the network interface name length. (The length does not include the terminating null byte).
	 * @param idxAddr
	 *            the index-position of the address in the interface addresses list
	 * @param addrInfo
	 *            the buffer into which the address information will be stored
	 * @param addrInfoLength
	 *            the length of the address information buffer.
	 * @return the network interface address size in bytes ({@link NetworkInterfaceSupport#IPV4_ADDR_INFO_SIZE
	 *         IPV4_ADDR_INFO_SIZE} for IPv4 or {@link NetworkInterfaceSupport#IPV6_ADDR_INFO_SIZE IPV6_ADDR_INFO_SIZE}
	 *         for IPv6).
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native int getVMInterfaceAddress(int idxIf, byte[] ifname, int ifnameLength, int idxAddr,
			byte[] addrInfo, int addrInfoLength) throws NativeIOException;

	/**
	 * Returns the number of network interfaces.
	 *
	 * @return the number of network interfaces.
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native int getVMInterfacesCount() throws NativeIOException;

	/**
	 * Checks if the given network interface {@code ifname} is a loopback interface.
	 *
	 * @param ifname
	 *            the network interface name (null-terminated string).
	 * @param ifnameLength
	 *            the network interface name length. (The length does not include the terminating null byte).
	 * @return 0 if true, 1 if false.
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native int isLoopback(byte[] ifname, int ifnameLength) throws NativeIOException;

	/**
	 * Checks if the given network interface {@code ifname} is a point-to-point interface.
	 *
	 * @param ifname
	 *            the network interface name (null-terminated string).
	 * @param ifnameLength
	 *            the network interface name length. (The length does not include the terminating null byte).
	 * @return 0 if true, 1 if false.
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native int isPointToPoint(byte[] ifname, int ifnameLength) throws NativeIOException;

	/**
	 * Checks if the given network interface {@code ifname} is up.
	 *
	 * @param ifname
	 *            the network interface name (null-terminated string).
	 * @param ifnameLength
	 *            the network interface name length. (The length does not include the terminating null byte).
	 * @return 0 if true, 1 if false.
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native int isUp(byte[] ifname, int ifnameLength) throws NativeIOException;

	/**
	 * Checks if the given network interface {@code ifname} supports multicast.
	 *
	 * @param ifname
	 *            the network interface name (null-terminated string).
	 * @param ifnameLength
	 *            the network interface name length. (The length does not include the terminating null byte).
	 * @return 0 if true, 1 if false.
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native int supportsMulticast(byte[] ifname, int ifnameLength) throws NativeIOException;

	/**
	 * Gets the the Maximum Transmission Unit (MTU) of the interface with the given network interface name
	 * {@code ifname}.
	 *
	 * @param ifname
	 *            the network interface name buffer (null-terminated string).
	 * @param ifnameLength
	 *            the network interface name length. (The length does not include the terminating null byte).
	 * @return the MTU as an int.
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native int getMTU(byte[] ifname, int ifnameLength) throws NativeIOException;

}
