/*
 * Copyright (c) 2015, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2015-2022 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.is2t.support.net.ssl.natives;

import com.is2t.support.net.ssl.SSLErrorConstants;

import ej.sni.NativeIOException;

public class SSLContextSupportNatives {

	static {
		SSLSocketSupportNatives.initializeNativePart();
	}

	/**
	 * Creates a new client ssl context based on the given SSL/TLS protocol version.
	 *
	 * @param protocol
	 *            the SSL/TLS protocol version (0 for SSLv3, 1 for TLSv1, 2 for TLSv1.1, 3 for TLSv1.2, 4 for DTLSv1.0,
	 *            5 for DTLS v1.2, and 6 for TLSv1.3).
	 * @param isClientContext
	 *            controls whether the underlying context will be created to manage ssl client or ssl server sockets.
	 *            <code>true</code> for client sockets; <code>false</code> for server sockets.
	 * @return the new ssl context ID on success.
	 *
	 * @throws NativeIOException
	 *             if an error occurs.
	 *
	 * @see {@link SSLErrorConstants} for error codes.
	 */
	public static native int createContext(int protocol, boolean isClientContext) throws NativeIOException;

	/**
	 * Releases the ssl context identified by the given contextID.
	 *
	 * @param contextID
	 *            the ssl context ID.
	 *
	 * @throws NativeIOException
	 *             if an error occurs.
	 *
	 * @see {@link SSLErrorConstants} for error codes.
	 */
	public static native void freeContext(int contextID) throws NativeIOException;

	/**
	 * Adds the given certificate as trusted certificate to the ssl context.
	 *
	 * @param contextID
	 *            the ssl context ID.
	 * @param certificate
	 *            the certificate to be added.
	 * @param certificateSize
	 *            the certificate size (in bytes).
	 * @param format
	 *            the certificate encoded format (0 for PEM format and 1 for DER format).
	 *
	 * @throws NativeIOException
	 *             if an error occurs.
	 *
	 * @see {@link SSLErrorConstants} for error codes.
	 */
	public static native void addTrustedCertificate(int contextID, byte[] certificate, int certificateSize, int format)
			throws NativeIOException;

	/**
	 * Removes all trusted certificates from the ssl context.
	 *
	 * @param contextID
	 *            the ssl context ID.
	 *
	 * @throws NativeIOException
	 *             if an error occurs.
	 *
	 * @see {@link SSLErrorConstants} for error codes.
	 */
	public static native void clearTrustStore(int contextID) throws NativeIOException;

	/**
	 * Removes all credentials (private key and associated certificates) from the ssl context.
	 *
	 * @param contextID
	 *            the ssl context ID.
	 *
	 * @throws NativeIOException
	 *             if an error occurs.
	 *
	 * @see {@link SSLErrorConstants} for error codes.
	 */
	public static native void clearKeyStore(int contextID) throws NativeIOException;

	/**
	 * Sets the certificate to be used in the ssl context.
	 *
	 * @param contextID
	 *            the ssl context ID.
	 * @param certificate
	 *            the certificate of the private key.
	 * @param certificateSize
	 *            the certificate size (in bytes).
	 * @param format
	 *            the certificate encoded format (0 for PEM format and 1 for DER format).
	 *
	 * @throws NativeIOException
	 *             if an error occurs.
	 *
	 * @see {@link SSLErrorConstants} for error codes.
	 */
	public static native void setCertificate(int contextID, byte[] certificate, int certificateSize, int format)
			throws NativeIOException;

	/**
	 * Sets the private key to be used in the ssl context. The provided private key must be in DER-encoded encrypted
	 * PKCS#8 format.
	 *
	 * @param contextID
	 *            the ssl context ID.
	 * @param privateKey
	 *            the private key (in DER-encoded encrypted PKCS#8 format).
	 * @param privateKeySize
	 *            the private key size (in bytes).
	 * @param password
	 *            The password string to be used to recover the encrypted PKCS#8 private key. This password is a
	 *            null-terminated string.
	 * @param passwordLength
	 *            The length of the password (in bytes). The length does not include the terminating null byte.
	 *
	 * @throws NativeIOException
	 *             if an error occurs.
	 *
	 * @see {@link SSLErrorConstants} for error codes.
	 */
	public static native void setPrivateKey(int contextID, byte[] privateKey, int privateKeySize, byte[] password,
			int passwordLength) throws NativeIOException;

	/**
	 * Initializes the certificates chain buffer and returns its total size.
	 *
	 * @param contextID
	 *            the ssl context ID.
	 * @param nbChainCerts
	 *            the number of certificates that will be added in the chain.
	 * @param chainCertsTotalSize
	 *            the total size of certificates in the chain (in bytes).
	 * @return the total size of the chain buffer (in bytes).
	 *
	 * @throws NativeIOException
	 *             if an error occurs.
	 *
	 * @see {@link SSLErrorConstants} for error codes.
	 */
	public static native int initChainBuffer(int contextID, int nbChainCerts, int chainCertsTotalSize)
			throws NativeIOException;

	/**
	 * Adds the given certificate to the certificates's chain of the ssl context.
	 *
	 * @param contextID
	 *            the ssl context ID.
	 * @param certificate
	 *            the certificate to be added in the chain.
	 * @param certificateSize
	 *            the certificate size (in bytes).
	 * @param format
	 *            the certificate encoded format (0 for PEM format and 1 for DER format).
	 * @param chainBufferSize
	 *            the size of the chain buffer (in bytes).
	 *
	 * @throws NativeIOException
	 *             if an error occurs.
	 *
	 * @see {@link SSLErrorConstants} for error codes.
	 */
	public static native void addChainCertificate(int contextID, byte[] certificate, int certificateSize, int format,
			int chainBufferSize) throws NativeIOException;

}
