/*
 * Copyright (c) 2015, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2015-2022 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.is2t.support.net.ssl.natives;

import com.is2t.support.net.ssl.SSLErrorConstants;
import com.is2t.support.net.ssl.util.NativeIOExceptionHandler;

import ej.sni.NativeIOException;

public class SSLSocketSupportNatives {

	static {
		initializeNativePart();
	}

	private static boolean initialized = false;

	/**
	 * Initializes SSL/TLS stack components
	 */
	public static native void initialize() throws NativeIOException;

	/**
	 * Creates a new ssl socket using the given ssl context and the underlying socket file descriptor {@code fd} for I/O
	 * operations.
	 *
	 * @param contextID
	 *            the ssl context ID.
	 * @param fd
	 *            the underlying socket file descriptor used for I/O operations.
	 * @param hostname
	 *            the server hostname (null-terminated string). The hostname is <code>null</code> if the server does not
	 *            have a valid domain name.
	 * @param hostnameLength
	 *            the length of the server hostname. (The length does not include the terminating null byte). The length
	 *            is <code>0</code> if the server does not have a valid domain name.
	 * @param autoclose
	 *            true if the underlying socket will be closed when this ssl socket is closed; false otherwise.
	 * @param useClientMode
	 *            configures the socket to use client (or server) mode during the handshake. <code>true</code> for
	 *            client mode; <code>false</code> for server mode.
	 * @param needClientAuth
	 *            configures the socket to require or not the client authentication. This option is only useful for
	 *            sockets in the server mode.<code>true</code> if the client authentication is required;
	 *            <code>false</code> otherwise.
	 * @return the new ssl socket ID on success.
	 *
	 * @throws NativeIOException
	 *             if an error occurs.
	 *
	 * @see {@link SSLErrorConstants} for error codes.
	 */
	public static native int create(int contextID, int fd, byte[] hostname, int hostnameLength, boolean autoclose,
			boolean useClientMode, boolean needClientAuth) throws NativeIOException;

	/**
	 * Shuts down the SSL/TLS connection.
	 *
	 * This method will try to send a "close notify" alert to the peer and then wait for the peer to send its "close
	 * notify" if <code>autoclose</code> is <code>false</code>. Waiting for the peer to send its "close notify" will
	 * consist of reading until the received data is a "close notify" alert.
	 *
	 * If <code>autoclose</code> is <code>true</code>, there is no need to wait for the peer to send its "close notify"
	 * since the underlying TCP connection will be closed right after this method call.
	 *
	 * @param sslID
	 *            the ssl socket ID.
	 * @param fd
	 *            the associated underlying socket file descriptor.
	 * @param autoclose
	 *            if <code>true</code>, no need to wait for the peer to send its "close notify"; otherwise we need to
	 *            read until the received data is a "close notify" alert.
	 * @param absoluteJavaStartTime
	 *            the absolute start time of the call of this method. This time is computed from the system time
	 *            returned by <B>ej.bon.Util.platformTimeMillis()</B>.
	 * @param relativeTimeout
	 *            the relative timeout in milliseconds for each underlying TCP read operation performed to read the
	 *            "close notify" alert; or 0 if no timeout.
	 *
	 * @throws NativeIOException
	 *             if an error occurs.
	 *
	 * @see {@link SSLErrorConstants} for error codes.
	 */
	public static native void shutdown(int sslID, int fd, boolean autoclose, long absoluteJavaStartTime,
			int relativeTimeout) throws NativeIOException;

	/**
	 * Releases the ssl socket identified by the given ssl ID.
	 *
	 * @param sslID
	 *            the ssl socket ID.
	 *
	 * @throws NativeIOException
	 *             if an error occurs.
	 *
	 * @see {@link SSLErrorConstants} for error codes.
	 */
	public static native void freeSSL(int sslID) throws NativeIOException;

	/**
	 * Initiates the handshake in client mode.
	 *
	 * @param sslID
	 *            the ssl socket ID.
	 * @param fd
	 *            the associated underlying socket file descriptor.
	 * @param absoluteJavaStartTime
	 *            the absolute start time of the call of this method. This time is computed from the system time
	 *            returned by <B>ej.bon.Util.platformTimeMillis()</B>.
	 * @param relativeTimeout
	 *            the relative timeout in milliseconds for each underlying read operation performed during the
	 *            handshake; or 0 if no timeout.
	 *
	 * @throws NativeIOException
	 *             if an error occurs.
	 *
	 * @see {@link SSLErrorConstants} for error codes.
	 */
	public static native void initialClientHandShake(int sslID, int fd, long absoluteJavaStartTime, int relativeTimeout)
			throws NativeIOException;

	/**
	 * Initiates the handshake in server mode.
	 *
	 * @param sslID
	 *            the ssl socket ID.
	 * @param fd
	 *            the associated underlying socket file descriptor.
	 * @param absoluteJavaStartTime
	 *            the absolute start time of the call of this method. This time is computed from the system time
	 *            returned by <B>ej.bon.Util.platformTimeMillis()</B>.
	 * @param relativeTimeout
	 *            the relative timeout in milliseconds for each underlying read operation performed during the
	 *            handshake; or 0 if no timeout.
	 *
	 * @throws NativeIOException
	 *             if an error occurs.
	 *
	 * @see {@link SSLErrorConstants} for error codes.
	 *
	 */
	public static native void initialServerHandShake(int sslID, int fd, long absoluteJavaStartTime, int relativeTimeout)
			throws NativeIOException;

	/**
	 * Reads {@code length} bytes from the ssl socket internal read buffer into the specified buffer {@code buffer}.
	 *
	 * @param sslID
	 *            the ssl socket ID.
	 * @param fd
	 *            the associated underlying socket file descriptor.
	 * @param buffer
	 *            the buffer into which the data is read.
	 * @param offset
	 *            the start offset in the buffer at which the data is written.
	 * @param length
	 *            the maximum number of data to read.
	 * @param absoluteJavaStartTime
	 *            the absolute start time of the call of this method. This time is computed from the system time
	 *            returned by <B>ej.bon.Util.platformTimeMillis()</B>.
	 * @param relativeTimeout
	 *            the relative timeout in milliseconds or 0 if no timeout.
	 * @return the number of bytes read on success or {@link SSLErrorConstants#J_EOF} if there is no more data because
	 *         the end of the stream has been reached.
	 *
	 * @throws NativeIOException
	 *             if an error occurs.
	 *
	 * @see {@link SSLErrorConstants} for error codes.
	 */
	public static native int read(int sslID, int fd, byte[] buffer, int offset, int length, long absoluteJavaStartTime,
			int relativeTimeout) throws NativeIOException;

	/**
	 * Returns the number of bytes available in the ssl socket internal read buffer.
	 *
	 * @param sslID
	 *            the ssl socket ID.
	 * @return the number of available bytes.
	 *
	 * @throws NativeIOException
	 *             if an error occurs.
	 *
	 * @see {@link SSLErrorConstants} for error codes.
	 */
	public static native int available(int sslID) throws NativeIOException;

	/**
	 * Writes {@code length} bytes from the specified buffer {@code buffer} starting at the offset to the ssl socket
	 * output stream. As a call to write operation may cause read operations (when a re-negotiation is requested for
	 * example), the given relative timeout will be used for these read operations.
	 *
	 * @param sslID
	 *            the ssl socket ID.
	 * @param fd
	 *            the associated underlying socket file descriptor.
	 * @param buffer
	 *            the buffer of data to write.
	 * @param offset
	 *            the offset in the buffer.
	 * @param length
	 *            the number of data to write.
	 * @param absoluteJavaStartTime
	 *            the absolute start time of the call if this method. This time is computed from the system time
	 *            returned by <B>ej.bon.Util.platformTimeMillis()</B>.
	 * @param relativeTimeout
	 *            the relative timeout in milliseconds for each underlying read operation performed during the write
	 *            operation; or 0 if no timeout.
	 *
	 * @return the number of bytes written on success.
	 *
	 * @throws NativeIOException
	 *             if an error occurs.
	 *
	 * @see {@link SSLErrorConstants} for error codes.
	 */
	public static native int write(int sslID, int fd, byte[] buffer, int offset, int length, long absoluteJavaStartTime,
			int relativeTimeout) throws NativeIOException;

	/**
	 * Initialize the foundation library native part.
	 */
	public static void initializeNativePart() {
		if (!initialized) {
			try {
				SSLSocketSupportNatives.initialize();
			} catch (NativeIOException e) {
				throw NativeIOExceptionHandler.generateNativeRuntimeException(e);
			}
		}
		initialized = true;
	}
}
