/*
 * Copyright (c) 2015, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2015-2022 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.is2t.support.net.ssl.x509;

import java.security.Key;
import java.security.KeyManagementException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.UnrecoverableKeyException;
import java.security.cert.Certificate;
import java.security.cert.CertificateEncodingException;
import java.util.Enumeration;

import javax.net.ssl.X509KeyManager;

import com.is2t.support.net.ssl.natives.SSLContextSupportNatives;
import com.is2t.support.net.ssl.util.NativeIOExceptionHandler;
import com.is2t.support.security.x509.X509CertImpl;
import com.is2t.support.security.x509.X509KeyImpl;

import ej.sni.NativeIOException;
import ej.sni.SNI;

public class X509KeyManagerImpl implements X509KeyManager {

	private X509KeyImpl privKey;
	private X509CertImpl[] certificates;
	private String password;

	public X509KeyManagerImpl(KeyStore ks, char[] password)
			throws KeyStoreException, UnrecoverableKeyException, NoSuchAlgorithmException {
		privKey = null;
		certificates = null;
		this.password = null;
		// credentialsMap = new HashMap<String, X509Credentials>();
		if (ks == null) {
			return;
		}

		// search the first credentials whose key and certificate are
		// respectively an implementation of X509KeyImpl and X509CertImpl
		for (Enumeration<String> aliases = ks.aliases(); aliases.hasMoreElements();) {
			String alias = aliases.nextElement();
			if (!ks.isKeyEntry(alias)) {
				continue;
			}
			Key key = ks.getKey(alias, password);
			if (!(key instanceof X509KeyImpl)) {
				continue;
			}
			Certificate[] certs = ks.getCertificateChain(alias);
			if ((certs == null) || (certs.length == 0) || !(certs[0] instanceof X509CertImpl)) {
				continue;
			}
			if (!(certs instanceof X509CertImpl[])) {
				Certificate[] tmp = new X509CertImpl[certs.length];
				// TODO be sure all certificates are instanceof X509CertImpl
				// before copy!
				System.arraycopy(certs, 0, tmp, 0, certs.length);
				certs = tmp;
			}
			privKey = (X509KeyImpl) key;
			certificates = (X509CertImpl[]) certs;
			if (password != null) {
				this.password = new String(password);
			}
			return;
		}
	}

	public void initContextKeyStore(int contextID) throws KeyManagementException, CertificateEncodingException {
		// clear the context key store
		try {
			SSLContextSupportNatives.clearKeyStore(contextID);
		} catch (NativeIOException e) {
			throw NativeIOExceptionHandler.generateNativeKeyManagementException(e);
		}
		// init with the private key and the associated chain certificates
		useCredentials(contextID, privKey, password, certificates);
	}

	private static void useCredentials(int contextID, X509KeyImpl key, String password, X509CertImpl[] certs)
			throws KeyManagementException, CertificateEncodingException {
		try {
			if (key != null && certs != null && certs.length > 0) {
				// on first, set the certificate associated with the private key
				// this cert is at index 0
				int format = certs[0].getEncodingFormat().ordinal();
				byte[] certByteArray = certs[0].getEncoded();

				SSLContextSupportNatives.setCertificate(contextID, certByteArray, certByteArray.length, format);

				// set the private key
				int passwordLength = password.getBytes().length;
				byte[] passwordByteArray = SNI.toCString(password);
				SSLContextSupportNatives.setPrivateKey(contextID, key.getKey(), key.getKeyLength(), passwordByteArray,
						passwordLength);

				// compute the total size of the rest of the certificate (the
				// issuers chain certificates)
				int chainCertsTotalSize = 0;
				int nbChainCerts = certs.length - 1;
				for (int i = 1; i <= nbChainCerts; i++) {
					chainCertsTotalSize += certs[i].getEncoded().length;
				}

				// init the chain buffer
				int chainBufferSize = SSLContextSupportNatives.initChainBuffer(contextID, nbChainCerts,
						chainCertsTotalSize);
				// set the certificate chain if any
				for (int i = 1; i <= nbChainCerts; i++) {
					format = certs[i].getEncodingFormat().ordinal();
					certByteArray = certs[i].getEncoded();
					SSLContextSupportNatives.addChainCertificate(contextID, certByteArray, certByteArray.length, format,
							chainBufferSize);
				}
			}
		} catch (NativeIOException e) {
			throw NativeIOExceptionHandler.generateNativeKeyManagementException(e);
		}
	}
}
