/*
 * Copyright (c) 2012, 2023, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2024 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package com.is2t.support.security.cipher;

import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.Key;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.security.spec.AlgorithmParameterSpec;
import java.security.spec.InvalidKeySpecException;

import javax.crypto.BadPaddingException;
import javax.crypto.CipherSpi;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.SecretKey;
import javax.crypto.ShortBufferException;
import javax.crypto.spec.PBEKeySpec;
import javax.crypto.spec.SecretKeySpec;

import com.microej.security.util.Arrays;
import com.microej.support.security.secretkey.NativeSecretKeyFactorySpi;

import sun.security.util.PBEUtil;

/**
 * Implementation copied from <code>com.sun.crypto.provider.PBES2Core</code>.
 */

/**
 * This class represents password-based encryption as defined by the PKCS #5 standard. These algorithms implement PBE
 * with HmacSHA1/HmacSHA2-family and AES-CBC. Padding is done as described in PKCS #5.
 *
 * @author Jan Luehe
 *
 *
 * @see javax.crypto.Cipher
 */
abstract class AbstractPBES2CipherSpi extends CipherSpi {
	// the encapsulated cipher
	private final NativeCipherSpi cipher;
	private final int keyLength; // in bits
	private final int blkSize; // in bytes
	private final NativeSecretKeyFactorySpi skf;
	private final String cipherAlgo;
	private final PBEUtil.PBES2Params pbes2Params = new PBEUtil.PBES2Params();

	/**
	 * Creates an instance of PBE Scheme 2 according to the selected password-based key derivation function and
	 * encryption scheme.
	 */
	AbstractPBES2CipherSpi(NativeSecretKeyFactorySpi secretKeyFactory, NativeCipherSpi cipher)
			throws NoSuchAlgorithmException, NoSuchPaddingException {
		this.skf = secretKeyFactory;
		this.cipher = cipher;

		this.cipherAlgo = "AES";
		this.blkSize = this.cipher.engineGetBlockSize();
		this.keyLength = this.blkSize * 8;// only AES_128 supported, so the key length is equal to the AES block size

		this.cipher.setMode("CBC");
		this.cipher.setPadding("PKCS5Padding");
	}

	@Override
	protected int engineGetBlockSize() {
		return this.blkSize;
	}

	@Override
	protected int engineGetOutputSize(int inputLen) {
		return this.cipher.engineGetOutputSize(inputLen);
	}

	@Override
	protected byte[] engineGetIV() {
		return this.cipher.engineGetIV();
	}

	@Override
	protected void engineInit(int opmode, Key key, SecureRandom random) throws InvalidKeyException {
		try {
			engineInit(opmode, key, (AlgorithmParameterSpec) null, random);
		} catch (InvalidAlgorithmParameterException ie) {
			throw new InvalidKeyException("requires PBE parameters", ie);
		}
	}

	@Override
	protected void engineInit(int opmode, Key key, AlgorithmParameterSpec params, SecureRandom random)
			throws InvalidKeyException, InvalidAlgorithmParameterException {
		PBEKeySpec pbeSpec = this.pbes2Params.getPBEKeySpec(this.blkSize, this.keyLength, opmode, key, params, random);
		SecretKey s = null;
		byte[] derivedKey;
		try {
			s = this.skf.engineGenerateSecret(pbeSpec);
			derivedKey = s.getEncoded();
		} catch (InvalidKeySpecException ikse) {
			throw new InvalidKeyException("Cannot construct PBE key", ikse);
		} finally {
			pbeSpec.clearPassword();
		}

		SecretKeySpec cipherKey = null;
		try {
			cipherKey = new SecretKeySpec(derivedKey, this.cipherAlgo);
			// initialize the underlying cipher
			this.cipher.engineInit(opmode, cipherKey, this.pbes2Params.getIvSpec(), random);
		} finally {
			Arrays.fill(derivedKey, (byte) 0);
		}
	}

	@Override
	protected byte[] engineUpdate(byte[] input, int inputOffset, int inputLen) {
		return this.cipher.engineUpdate(input, inputOffset, inputLen);
	}

	@Override
	protected int engineUpdate(byte[] input, int inputOffset, int inputLen, byte[] output, int outputOffset)
			throws ShortBufferException {
		return this.cipher.engineUpdate(input, inputOffset, inputLen, output, outputOffset);
	}

	@Override
	protected byte[] engineDoFinal(byte[] input, int inputOffset, int inputLen)
			throws IllegalBlockSizeException, BadPaddingException {
		return this.cipher.engineDoFinal(input, inputOffset, inputLen);
	}

	@Override
	protected int engineDoFinal(byte[] input, int inputOffset, int inputLen, byte[] output, int outputOffset)
			throws ShortBufferException, IllegalBlockSizeException, BadPaddingException {
		return this.cipher.engineDoFinal(input, inputOffset, inputLen, output, outputOffset);
	}

}
