/*
 * Copyright (c) 2002, 2023, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2024 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package com.is2t.support.security.cipher;

import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.Key;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.security.spec.AlgorithmParameterSpec;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.CipherSpi;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.SecretKey;
import javax.crypto.ShortBufferException;
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.spec.PBEParameterSpec;
import javax.crypto.spec.SecretKeySpec;

import com.is2t.support.security.digest.NativeMessageDigestSpi;
import com.microej.security.util.Arrays;

/**
 * Implementation copied from <code>com.sun.crypto.provider.PKCS12PBECipherCore</code>.
 */

/**
 * This class represents password-based encryption as defined by the PKCS #5 standard.
 *
 * @author Jan Luehe
 *
 *
 * @see javax.crypto.Cipher
 */
abstract class AbstractPKCS12PBECipherSpi extends CipherSpi {

	private static final int DES_EDE_KEY_SIZE_BITS = 192;
	private static final int DES_BLOCK_SIZE = 8;

	private static final int DEFAULT_SALT_LENGTH = 20;
	private static final int DEFAULT_COUNT = 1024;

	static final int CIPHER_KEY = 1;
	static final int CIPHER_IV = 2;

	// the encapsulated DES cipher
	private final NativeCipherSpi cipher;
	private final NativeMessageDigestSpi md;
	private final String algo;
	private byte[] salt = null;
	private int iCount = 10;
	private final int keySize; // in bytes

	/**
	 * Creates an instance of PBE Cipher using the specified CipherSpi instance.
	 *
	 */
	AbstractPKCS12PBECipherSpi(NativeMessageDigestSpi md, NativeCipherSpi cipher)
			throws NoSuchAlgorithmException, NoSuchPaddingException {
		this.md = md;
		this.cipher = cipher;
		this.keySize = DES_EDE_KEY_SIZE_BITS >> 3;

		this.algo = "DESede";
		this.cipher.setMode("CBC");
		this.cipher.setPadding("PKCS5Padding");
	}

	@Override
	protected int engineGetBlockSize() {
		return DES_BLOCK_SIZE;
	}

	@Override
	protected int engineGetOutputSize(int inputLen) {
		return this.cipher.engineGetOutputSize(inputLen);
	}

	@Override
	protected byte[] engineGetIV() {
		return this.cipher.engineGetIV();
	}

	@Override
	protected void engineInit(int opmode, Key key, SecureRandom random) throws InvalidKeyException {
		try {
			engineInit(opmode, key, (AlgorithmParameterSpec) null, random);
		} catch (InvalidAlgorithmParameterException ie) {
			throw new InvalidKeyException("requires PBE parameters", ie);
		}
	}

	@Override
	protected void engineInit(int opmode, Key key, AlgorithmParameterSpec params, SecureRandom random)
			throws InvalidKeyException, InvalidAlgorithmParameterException {
		char[] passwdChars = null;
		this.salt = null;
		this.iCount = 0;
		if (key instanceof javax.crypto.interfaces.PBEKey) {
			javax.crypto.interfaces.PBEKey pbeKey = (javax.crypto.interfaces.PBEKey) key;
			passwdChars = pbeKey.getPassword();
			this.salt = pbeKey.getSalt(); // maybe null if unspecified
			this.iCount = pbeKey.getIterationCount(); // maybe 0 if unspecified
		} else if (key instanceof SecretKey) {
			byte[] passwdBytes;
			if (!(key.getAlgorithm().regionMatches(true, 0, "PBE", 0, 3)) || (passwdBytes = key.getEncoded()) == null) {
				throw new InvalidKeyException("Missing password");
			}
			passwdChars = new char[passwdBytes.length];
			for (int i = 0; i < passwdChars.length; i++) {
				passwdChars[i] = (char) (passwdBytes[i] & 0x7f);
			}
			Arrays.fill(passwdBytes, (byte) 0x00);
		} else {
			throw new InvalidKeyException("SecretKey of PBE type required");
		}

		try {
			if (((opmode == Cipher.DECRYPT_MODE) || (opmode == Cipher.UNWRAP_MODE))
					&& ((params == null) && ((this.salt == null) || (this.iCount == 0)))) {
				throw new InvalidAlgorithmParameterException("Parameters missing");
			}

			if (params == null) {
				// generate default for salt and iteration count if necessary
				if (this.salt == null) {
					this.salt = new byte[DEFAULT_SALT_LENGTH];
					random.nextBytes(this.salt);
				}
				if (this.iCount == 0) {
					this.iCount = DEFAULT_COUNT;
				}
			} else if (!(params instanceof PBEParameterSpec)) {
				throw new InvalidAlgorithmParameterException("PBEParameterSpec type required");
			} else {
				PBEParameterSpec pbeParams = (PBEParameterSpec) params;
				// make sure the parameter values are consistent
				if (this.salt != null) {
					if (!Arrays.equals(this.salt, pbeParams.getSalt())) {
						throw new InvalidAlgorithmParameterException(
								"Inconsistent value of salt between key and params");
					}
				} else {
					this.salt = pbeParams.getSalt();
				}
				if (this.iCount != 0) {
					if (this.iCount != pbeParams.getIterationCount()) {
						throw new InvalidAlgorithmParameterException(
								"Different iteration count between key and params");
					}
				} else {
					this.iCount = pbeParams.getIterationCount();
				}
			}
			// salt is recommended to be ideally as long as the output
			// of the hash function. However, it may be too strict to
			// force this; so instead, we'll just require the minimum
			// salt length to be 8-byte which is what PKCS#5 recommends
			// and openssl does.
			if (this.salt.length < 8) {
				throw new InvalidAlgorithmParameterException("Salt must be at least 8 bytes long");
			}
			if (this.iCount <= 0) {
				throw new InvalidAlgorithmParameterException("IterationCount must be a positive number");
			}

			byte[] derivedKey = derive(passwdChars, this.salt, this.iCount, this.keySize, CIPHER_KEY, this.md,
					DES_BLOCK_SIZE * 8);
			SecretKey cipherKey = new SecretKeySpec(derivedKey, this.algo);
			Arrays.fill(derivedKey, (byte) 0);

			// initialize the underlying cipher
			this.cipher.engineInit(opmode, cipherKey,
					new IvParameterSpec(
							derive(passwdChars, this.salt, this.iCount, 8, CIPHER_IV, this.md, DES_BLOCK_SIZE * 8)),
					random);
		} finally {
			Arrays.fill(passwdChars, '\0');
		}
	}

	// Uses supplied hash algorithm
	static byte[] derive(char[] chars, byte[] salt, int ic, int n, int type, NativeMessageDigestSpi md,
			int blockLength) {

		// Add in trailing NULL terminator. Special case:
		// no terminator if password is "\0".
		int length = chars.length * 2;
		if (length == 2 && chars[0] == 0) {
			chars = new char[0];
			length = 0;
		} else {
			length += 2;
		}

		byte[] passwd = new byte[length];
		for (int i = 0, j = 0; i < chars.length; i++, j += 2) {
			passwd[j] = (byte) ((chars[i] >>> 8) & 0xFF);
			passwd[j + 1] = (byte) (chars[i] & 0xFF);
		}
		byte[] key = new byte[n];

		try {
			int v = blockLength;
			int u = md.engineGetDigestLength();
			int c = roundup(n, u) / u;
			byte[] D = new byte[v];
			int s = roundup(salt.length, v);
			int p = roundup(passwd.length, v);
			byte[] I = new byte[s + p];

			Arrays.fill(D, (byte) type);
			concat(salt, I, 0, s);
			concat(passwd, I, s, p);
			Arrays.fill(passwd, (byte) 0x00);

			byte[] Ai;
			byte[] B = new byte[v];

			int i = 0;
			for (;; i++, n -= u) {
				md.engineUpdate(D, 0, D.length);
				md.engineUpdate(I, 0, I.length);
				Ai = md.engineDigest();
				for (int r = 1; r < ic; r++) {
					md.engineUpdate(Ai, 0, Ai.length);
					Ai = md.engineDigest();
				}
				System.arraycopy(Ai, 0, key, u * i, Math.min(n, u));
				if (i + 1 == c) {
					break;
				}
				concat(Ai, B, 0, v);
				addOne(v, B); // add 1 into B

				for (int j = 0; j < I.length; j += v) {
					addTwo(v, B, I, j); // add B into I from j
				}
			}
			Arrays.fill(I, (byte) 0);
		} catch (Exception e) {
			throw new RuntimeException("internal error: " + e);
		}
		return key;
	}

	// Add 1 to b (as integer)
	private static void addOne(int len, byte[] b) {
		for (int i = len - 1; i >= 0; i--) {
			if ((b[i] & 0xff) != 255) {
				b[i]++;
				break;
			} else {
				b[i] = 0;
			}
		}
	}

	// Add src (as integer) to dst from offset (as integer)
	private static void addTwo(int len, byte[] src, byte[] dst, int offset) {
		int carry = 0;
		for (int i = len - 1; i >= 0; i--) {
			int sum = (src[i] & 0xff) + (dst[i + offset] & 0xff) + carry;
			carry = sum >> 8;
			dst[i + offset] = (byte) sum;
		}
	}

	private static int roundup(int x, int y) {
		return ((x + (y - 1)) / y) * y;
	}

	private static void concat(byte[] src, byte[] dst, int start, int len) {
		if (src.length == 0) {
			return;
		}
		int loop = len / src.length;
		int off, i;
		for (i = 0, off = 0; i < loop; i++, off += src.length) {
			System.arraycopy(src, 0, dst, off + start, src.length);
		}
		System.arraycopy(src, 0, dst, off + start, len - off);
	}

	@Override
	protected byte[] engineUpdate(byte[] input, int inputOffset, int inputLen) {
		return this.cipher.engineUpdate(input, inputOffset, inputLen);
	}

	@Override
	protected int engineUpdate(byte[] input, int inputOffset, int inputLen, byte[] output, int outputOffset)
			throws ShortBufferException {
		return this.cipher.engineUpdate(input, inputOffset, inputLen, output, outputOffset);
	}

	@Override
	protected byte[] engineDoFinal(byte[] input, int inputOffset, int inputLen)
			throws IllegalBlockSizeException, BadPaddingException {
		return this.cipher.engineDoFinal(input, inputOffset, inputLen);
	}

	@Override
	protected int engineDoFinal(byte[] input, int inputOffset, int inputLen, byte[] output, int outputOffset)
			throws ShortBufferException, IllegalBlockSizeException, BadPaddingException {
		return this.cipher.engineDoFinal(input, inputOffset, inputLen, output, outputOffset);
	}

}
