/*
 * Copyright (c) 1997, 2013, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2023-2024 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.is2t.support.security.cipher;

import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.Key;
import java.security.PublicKey;
import java.security.SecureRandom;
import java.security.spec.AlgorithmParameterSpec;
import java.security.spec.MGF1ParameterSpec;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.CipherSpi;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.ShortBufferException;
import javax.crypto.spec.OAEPParameterSpec;
import javax.crypto.spec.PSource;

import com.is2t.support.security.NativePrivateKey;
import com.is2t.support.security.NativePublicKey;

/**
 * Implementation copied from <code>com.sun.crypto.provider.RSACipher</code>.
 */

/**
 * RSA cipher implementation. Supports RSA en/decryption and signing/verifying using both PKCS#1 v1.5 and OAEP (v2.2)
 * paddings and without padding (raw RSA). Note that raw RSA is supported mostly for completeness and should only be
 * used in rare cases.
 *
 * Objects should be instantiated by calling Cipher.getInstance() using the following algorithm names: .
 * "RSA/ECB/PKCS1Padding" (or "RSA") for PKCS#1 v1.5 padding. . "RSA/ECB/OAEPwith<hash>andMGF1Padding" (or
 * "RSA/ECB/OAEPPadding") for PKCS#1 v2.2 padding. . "RSA/ECB/NoPadding" for rsa RSA.
 *
 * We only do one RSA operation per doFinal() call. If the application passes more data via calls to update() or
 * doFinal(), we throw an IllegalBlockSizeException when doFinal() is called (see JCE API spec). Bulk encryption using
 * RSA does not make sense and is not standardized.
 *
 * Note: RSA keys should be at least 512 bits long
 *
 * @since 1.5
 * @author Andreas Sterbenz
 */
public abstract class AbstractRSACipherSpi extends CipherSpi {

	// constant for an empty byte array
	private static final byte[] B0 = new byte[0];

	// mode constant for public key encryption
	private static final int MODE_ENCRYPT = 1;

	// mode constant for private key decryption
	private static final int MODE_DECRYPT = 2;

	/*
	 * Padding type constants. Native known.
	 */
	private static final int PAD_PKCS1_TYPE = 0;
	private static final int PAD_OAEP_MGF1_TYPE = 1;

	/*
	 * OAEP hash algorithm constants. Native known.
	 */
	private static final int OAEP_HASH_SHA_1_ALGORITHM = 0;
	private static final int OAEP_HASH_SHA_256_ALGORITHM = 1;

	// size in bytes of the PKCS1 padding
	private static final int PADDING_PKCS1_SIZE = 11;

	// size in bytes of the OAEP SHA-1 padding
	private static final int PADDING_OAEP_SHA_1_SIZE = 2 * 20;

	// size in bytes of the OAEP SHA-256 padding
	private static final int PADDING_OAEP_SHA_256_SIZE = 2 * 32;

	// current mode, one of MODE_* above. Set when init() is called
	private int mode;

	// active padding type, one of PAD_* above. Set by setPadding()
	private final int paddingType;

	// buffer for the data
	private byte[] buffer;

	// offset into the buffer (number of bytes buffered)
	private int bufOfs;

	// size of the output
	private int outputSize;

	// hash algorithm for OAEP
	private int oaepHashAlgorithm;

	/**
	 * Default constructor.
	 */
	public AbstractRSACipherSpi() {
		this(PAD_PKCS1_TYPE, 0);
	}

	/**
	 * Custom constructor.
	 *
	 * @param paddingType
	 *            one of PAD_* above.
	 * @param oaepHashAlgorithm
	 *            one of OAEP_HASH_* above.
	 */
	public AbstractRSACipherSpi(int paddingType, int oaepHashAlgorithm) {
		this.paddingType = paddingType;
		this.oaepHashAlgorithm = oaepHashAlgorithm;
	}

	@Override
	protected int engineGetBlockSize() {
		return 0;
	}

	@Override
	protected int engineGetOutputSize(int inputLen) {
		return this.outputSize;
	}

	@Override
	protected byte[] engineGetIV() {
		return null;
	}

	@Override
	protected void engineInit(int opmode, Key key, SecureRandom random) throws InvalidKeyException {
		try {
			init(opmode, key, random, null);
		} catch (InvalidAlgorithmParameterException iape) {
			InvalidKeyException ike = new InvalidKeyException("Wrong parameters");
			ike.initCause(iape);
			throw ike;
		}
	}

	@Override
	protected void engineInit(int opmode, Key key, AlgorithmParameterSpec params, SecureRandom random)
			throws InvalidKeyException, InvalidAlgorithmParameterException {
		init(opmode, key, random, params);
	}

	// initialize this cipher
	private void init(int opmode, Key key, SecureRandom random, AlgorithmParameterSpec params)
			throws InvalidAlgorithmParameterException {
		NativePublicKey publicKey = null;
		NativePrivateKey privateKey = null;
		boolean encrypt;
		AlgorithmParameterSpec spec = null;

		switch (opmode) {
		case Cipher.ENCRYPT_MODE:
			encrypt = true;
			break;
		case Cipher.DECRYPT_MODE:
			encrypt = false;
			break;
		default:
			throw new InvalidAlgorithmParameterException("Not supported mode: " + opmode);
		}

		if (key instanceof PublicKey) {
			if (encrypt) {
				this.mode = MODE_ENCRYPT;
			} else {
				throw new InvalidAlgorithmParameterException("Verify mode not supported");
			}
			publicKey = (NativePublicKey) key;
			this.outputSize = publicKey.getOutputSize();
		} else {
			if (!encrypt) {
				this.mode = MODE_DECRYPT;
			} else {
				throw new InvalidAlgorithmParameterException("Sign mode not supported");
			}
			privateKey = (NativePrivateKey) key;
			this.outputSize = privateKey.getOutputSize();
		}

		if (this.paddingType == PAD_PKCS1_TYPE) {
			if (encrypt) {
				this.buffer = new byte[this.outputSize - PADDING_PKCS1_SIZE];
			} else {
				this.buffer = new byte[this.outputSize];
			}
		} else if (this.paddingType == PAD_OAEP_MGF1_TYPE) {
			if (encrypt) {
				String oaepHashAlgo = "";

				if (this.oaepHashAlgorithm == OAEP_HASH_SHA_1_ALGORITHM) {
					oaepHashAlgo = "SHA-1";
					this.buffer = new byte[this.outputSize - PADDING_OAEP_SHA_1_SIZE - 2];
				} else if (this.oaepHashAlgorithm == OAEP_HASH_SHA_256_ALGORITHM) {
					oaepHashAlgo = "SHA-256";
					this.buffer = new byte[this.outputSize - PADDING_OAEP_SHA_256_SIZE - 2];
				} else {
					throw new InvalidAlgorithmParameterException(
							"OAEP Digest type not supported: " + this.oaepHashAlgorithm);
				}

				if (params != null) {
					if (!(params instanceof OAEPParameterSpec)) {
						throw new InvalidAlgorithmParameterException("Wrong Parameters for OAEP Padding");
					}
					spec = params;
				} else {
					spec = new OAEPParameterSpec(oaepHashAlgo, "MGF1", MGF1ParameterSpec.SHA1,
							PSource.PSpecified.DEFAULT);
				}
			} else {
				this.buffer = new byte[this.outputSize];
			}
		} else {
			throw new InvalidAlgorithmParameterException("Padding type not supported: " + this.paddingType);
		}

		if (spec != null) {
			MGF1ParameterSpec mgfSpec = (MGF1ParameterSpec) ((OAEPParameterSpec) spec).getMGFParameters();
			String mdName = mgfSpec.getDigestAlgorithm();
			if (mdName.equals("SHA-1")) {
				this.oaepHashAlgorithm = OAEP_HASH_SHA_1_ALGORITHM;
			} else if (mdName.equals("SHA-256")) {
				this.oaepHashAlgorithm = OAEP_HASH_SHA_256_ALGORITHM;
			} else {
				throw new InvalidAlgorithmParameterException("OAEP MGF1 spec not supported: " + mdName);
			}
		}

		init(publicKey, privateKey, this.paddingType, this.oaepHashAlgorithm);
	}

	// internal update method
	private void update(byte[] in, int inOfs, int inLen) {
		if ((inLen == 0) || (in == null)) {
			return;
		}

		if (inLen > (this.buffer.length - this.bufOfs)) {
			this.bufOfs = this.buffer.length + 1;
			return;
		}

		System.arraycopy(in, inOfs, this.buffer, this.bufOfs, inLen);
		this.bufOfs += inLen;
	}

	// internal doFinal() method. Here we perform the actual RSA operation
	private byte[] doFinal() throws IllegalBlockSizeException {
		byte[] data = new byte[this.outputSize];
		int size;

		if (this.bufOfs > this.buffer.length) {
			throw new IllegalBlockSizeException("Data must not be longer " + "than " + this.buffer.length + " bytes");
		}

		try {
			switch (this.mode) {
			case MODE_ENCRYPT:
				size = encrypt(this.buffer, 0, this.bufOfs, data, 0);
				break;
			case MODE_DECRYPT:
				size = decrypt(this.buffer, 0, this.bufOfs, data, 0);
				break;
			default:
				throw new AssertionError("Internal error");
			}
		} finally {
			this.bufOfs = 0;
		}

		byte[] output = new byte[size];
		System.arraycopy(data, 0, output, 0, size);

		return output;
	}

	@Override
	protected byte[] engineUpdate(byte[] in, int inOfs, int inLen) {
		update(in, inOfs, inLen);
		return B0;
	}

	@Override
	protected int engineUpdate(byte[] in, int inOfs, int inLen, byte[] out, int outOfs) {
		update(in, inOfs, inLen);
		return 0;
	}

	@Override
	protected byte[] engineDoFinal(byte[] in, int inOfs, int inLen)
			throws BadPaddingException, IllegalBlockSizeException {
		update(in, inOfs, inLen);
		return doFinal();
	}

	@Override
	protected int engineDoFinal(byte[] in, int inOfs, int inLen, byte[] out, int outOfs)
			throws ShortBufferException, BadPaddingException, IllegalBlockSizeException {
		if (this.outputSize > out.length - outOfs) {
			throw new ShortBufferException("Need " + this.outputSize + " bytes for output");
		}

		update(in, inOfs, inLen);
		byte[] result = doFinal();
		int n = result.length;
		System.arraycopy(result, 0, out, outOfs, n);
		return n;
	}

	/**
	 * Initialize the native part of the RSA cipher.
	 *
	 * @param publicKey
	 *            Public key of the RSA cipher (null for decryption).
	 * @param privateKey
	 *            Private key of the RSA cipher (null for encryption).
	 * @param padding
	 *            Padding scheme of the RSA cipher.
	 * @param hash
	 *            Hash algorithm of the RSA cipher (valid only for OAEP with MGF1 padding scheme).
	 */
	protected abstract void init(NativePublicKey publicKey, NativePrivateKey privateKey, int padding, int hash);

	/**
	 * Encrypt a buffer with the RSA cipher.
	 *
	 * @param in
	 *            input buffer
	 * @param inOffset
	 *            offset in the input buffer
	 * @param inLength
	 *            input buffer length
	 * @param output
	 *            output buffer
	 * @param outputOffset
	 *            offset in the output buffer length
	 * @return the size of the encrypted data
	 */
	protected abstract int encrypt(byte[] in, int inOffset, int inLength, byte[] output, int outputOffset);

	/**
	 * Decrypt a buffer with the RSA cipher.
	 *
	 * @param in
	 *            input buffer
	 * @param inOffset
	 *            offset in the input buffer
	 * @param inLength
	 *            input buffer length
	 * @param output
	 *            output buffer
	 * @param outputOffset
	 *            offset in the output buffer length
	 * @return the size of the decrypted data
	 */
	protected abstract int decrypt(byte[] in, int inOffset, int inLength, byte[] output, int outputOffset);

}
