/*
 * Copyright 2017-2024 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.is2t.support.security.cipher;

import java.security.InvalidKeyException;
import java.security.ProviderException;

import ej.bon.ByteArray;
import ej.sni.NativeException;
import ej.sni.NativeResource;

/**
 *
 */
public class NativeCipherSpi extends AbstractCipherSpi {

	/**
	 * @see #nativeGetTransformationDescription(byte[], byte[])
	 */
	public static final int CIPHER_DESCRIPTION_SIZE = ByteArray.INT_SIZE * 3;

	/**
	 * Native identifier for the cipher. Initialized during engine init.
	 */
	private int nativeId;

	private final int nativeTransformationId;

	public NativeCipherSpi(int nativeTransformationId, int blockSize, int unitBytes, int cipherMode) {
		super(blockSize, unitBytes, cipherMode);
		this.nativeId = -1;
		this.nativeTransformationId = nativeTransformationId;
	}

	@Override
	protected int getBufferedLength() {
		return nativeGetBufferedLength(this.nativeTransformationId, this.nativeId);
	}

	@Override
	protected byte[] engineGetIV() {
		int ivLength = nativeGetIVLength(this.nativeTransformationId, this.nativeId);
		if (ivLength < 0) {
			return null;
		}

		byte[] iv = new byte[ivLength];
		nativeGetIV(this.nativeTransformationId, this.nativeId, iv, ivLength);
		return iv;
	}

	@Override
	protected void engineSetIV(byte[] iv) {
		if (iv == null) {
			return;
		}

		int ivLength = iv.length;
		if (ivLength < 0) {
			return;
		}

		nativeSetIV(this.nativeTransformationId, this.nativeId, iv, ivLength);
	}

	@Override
	protected void engineInit(boolean decrypting, String algorithm, byte[] key, byte[] iv) throws InvalidKeyException {
		// TODO check key algorithm string value

		try {

			if (this.nativeId != -1) {
				// close and reset the existing resource
				nativeClose(this.nativeTransformationId, this.nativeId);
				this.nativeId = -1;
			}

			int nativeTransformationId = this.nativeTransformationId;
			try {
				this.nativeId = nativeInit(nativeTransformationId, decrypting, key, key.length, iv, iv.length);
				NativeResource.closeOnGC(this.nativeId, nativeGetCloseId(nativeTransformationId), this);

			} catch (NativeException e) {
				// An error occurs during initialization, maybe there is no space left to
				// allocate new native resource.
				// Trigger a new fresh GC and release new potential useless resources and try
				// the initialization again!
				System.gc();
				this.nativeId = nativeInit(nativeTransformationId, decrypting, key, key.length, iv, iv.length);
				NativeResource.closeOnGC(this.nativeId, nativeGetCloseId(nativeTransformationId), this);
			}

		} catch (NativeException e) {
			throw new ProviderException(e);
		}
	}

	@Override
	protected int decrypt(byte[] buffer, int i, int len, byte[] output, int outputOffset) {
		return nativeDecrypt(this.nativeTransformationId, this.nativeId, buffer, i, len, output, outputOffset);
	}

	@Override
	protected int encrypt(byte[] buffer, int i, int len, byte[] output, int outputOffset) {
		return nativeEncrypt(this.nativeTransformationId, this.nativeId, buffer, i, len, output, outputOffset);
	}

	/**
	 * Gets for the given transformation the cipher description.
	 * <p>
	 * <code>description</code> must be filled-in with:
	 * <ul>
	 * <li>[0-3]: blockSize: the block size (in bytes), or 0 if the underlying algorithm is not a block cipher</li>
	 * <li>[4-7]: unitBytes: unit size (number of input bytes that can be processed at a time)</li>
	 * <li>[8-11]: cipherMode: one of the <code>*_MODE</code> values defined in {@link AbstractCipherSpi}</li>
	 * </ul>
	 *
	 * @param transformation
	 *            Null terminated string that describes the transformation
	 * @return transformation ID on success or -1 on error.
	 */
	public static native int nativeGetTransformationDescription(byte[] transformation, byte[] description);

	/**
	 * Returns the number of bytes that are buffered internally inside the given cipher.
	 *
	 * @param nativeTransformationId
	 *            ID of the transformation
	 * @param nativeId
	 *            the resource's native ID
	 * @return the length of the buffer.
	 * @throws NativeException
	 *             on error
	 */
	private static native int nativeGetBufferedLength(int nativeTransformationId, int nativeId);

	/**
	 * Gets the initialization vector.
	 *
	 * @param nativeTransformationId
	 *            ID of the transformation
	 * @param nativeId
	 *            the resource's native ID
	 * @param iv
	 *            the initialization vector of the cipher
	 * @param ivLength
	 *            the initialization vector size
	 * @throws NativeException
	 *             on error
	 */
	private static native void nativeGetIV(int nativeTransformationId, int nativeId, byte[] iv, int ivLength);

	/**
	 * Sets the initialization vector.
	 *
	 * @param nativeTransformationId
	 *            ID of the transformation
	 * @param nativeId
	 *            the resource's native ID
	 * @param iv
	 *            the initialization vector of the cipher
	 * @param ivLength
	 *            the initialization vector size
	 * @throws NativeException
	 *             on error
	 */
	private static native void nativeSetIV(int nativeTransformationId, int nativeId, byte[] iv, int ivLength);

	/**
	 * Gets the length of the initialized vector.
	 *
	 * @param nativeTransformationId
	 *            ID of the transformation
	 * @param nativeId
	 *            the resource's native ID
	 * @return -1 if no IV, otherwise returns the length of the IV.
	 * @throws NativeException
	 *             on error
	 */
	private static native int nativeGetIVLength(int nativeTransformationId, int nativeId);

	/**
	 * Initializes a Cipher resource.
	 *
	 * @param nativeTransformationId
	 *            ID of the transformation
	 *
	 * @param decrypting
	 *            <code>true</code> for decrypting, <code>false</code> for encryting.
	 * @param key
	 *            the key
	 * @param key_length
	 *            the key size
	 * @param iv
	 *            the initialization vector
	 * @param iv_length
	 *            the initialization vector size
	 *
	 * @return the nativeId of the newly initialized resource
	 * @throws NativeException
	 *             on error.
	 */
	private static native int nativeInit(int nativeTransformationId, boolean decrypting, byte[] key, int keyLength,
			byte[] iv, int ivLength) throws NativeException;

	/**
	 * Decrypts the message contained in <code>buffer</code>.
	 *
	 * @param nativeTransformationId
	 *            ID of the transformation
	 * @param nativeId
	 *            the resource's native ID
	 * @param buffer
	 *            the buffer containing the message to decrypt
	 * @param bufferOffset
	 *            the buffer offset
	 * @param len
	 *            the buffer length
	 * @param output
	 *            the output buffer containing the plaintext message
	 * @param outputOffset
	 *            the output offset
	 * @return the length of the buffer
	 * @throws NativeException
	 *             on error
	 */
	private static native int nativeDecrypt(int nativeTransformationId, int nativeId, byte[] buffer, int bufferOffset,
			int len, byte[] output, int outputOffset) throws NativeException;

	/**
	 * Encrypts the message contained in <code>buffer</code>.
	 *
	 * @param nativeTransformationId
	 *            ID of the transformation
	 * @param nativeId
	 *            the resource's native ID
	 * @param buffer
	 *            the buffer containing the plaintext message to encrypt
	 * @param bufferOffset
	 *            the buffer offset
	 * @param len
	 *            the buffer length
	 * @param output
	 *            the output buffer containing the encrypted message
	 * @param outputOffset
	 *            the output offset
	 * @return the length of the buffer
	 * @throws NativeException
	 *             on error
	 */
	private static native int nativeEncrypt(int nativeTransformationId, int nativeId, byte[] buffer, int bufferOffset,
			int len, byte[] output, int outputOffset) throws NativeException;

	/**
	 * Closes the resource related to the native ID
	 *
	 * @param nativeTransformationId
	 *            ID of the transformation
	 * @param nativeId
	 *            the resource's native ID
	 * @throws NativeException
	 *             on error
	 */
	private static native void nativeClose(int nativeTransformationId, int nativeId) throws NativeException;

	/**
	 * Gets the id of the native close function.
	 *
	 * @param nativeTransformationId
	 *            ID of the transformation
	 *
	 * @return the id of the static native close function
	 * @throws NativeException
	 *             on error
	 */
	private static native int nativeGetCloseId(int nativeTransformationId) throws NativeException;

}
