/*
 * Copyright (c) 2003, 2014, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2017-2022 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.is2t.support.security.digest;

/**
 *
 */
abstract public class AbstractBufferedMessageDigestSpi extends AbstractMessageDigestSpi {

	// size of the input to the compression function in bytes
	private final int blockSize;

	// buffer to store partial blocks, blockSize bytes large
	// Subclasses should not access this array directly except possibly in their
	// implDigest() method. See MD5.java as an example.
	byte[] buffer;
	// offset into buffer
	private int bufOfs;

	AbstractBufferedMessageDigestSpi(int digestLength, int blockSize) {
		super(digestLength);
		this.blockSize = blockSize;
		this.buffer = new byte[blockSize];
	}

	@Override
	protected void engineReset() {
		super.engineReset();
		this.bufOfs = 0;
	}

	@Override
	void implUpdate(byte[] b, int ofs, int len) {
		this.bytesProcessed += len;
		// if buffer is not empty, we need to fill it before proceeding
		if (this.bufOfs != 0) {
			int n = Math.min(len, this.blockSize - this.bufOfs);
			System.arraycopy(b, ofs, this.buffer, this.bufOfs, n);
			this.bufOfs += n;
			ofs += n;
			len -= n;
			if (this.bufOfs >= this.blockSize) {
				// compress completed block now
				implCompress(this.buffer, 0);
				this.bufOfs = 0;
			}
		}
		// compress complete blocks
		if (len >= this.blockSize) {
			int limit = ofs + len;
			ofs = implCompressMultiBlock(b, ofs, limit - this.blockSize);
			len = limit - ofs;
		}
		// copy remainder to buffer
		if (len > 0) {
			System.arraycopy(b, ofs, this.buffer, 0, len);
			this.bufOfs = len;
		}
	}

	// compress complete blocks
	private int implCompressMultiBlock(byte[] b, int ofs, int limit) {
		for (; ofs <= limit; ofs += this.blockSize) {
			implCompress(b, ofs);
		}
		return ofs;
	}

	/**
	 * Core compression function. Processes blockSize bytes at a time and updates the state of this object.
	 */
	abstract void implCompress(byte[] b, int ofs);

	@Override
	public Object clone() throws CloneNotSupportedException {
		AbstractBufferedMessageDigestSpi copy = (AbstractBufferedMessageDigestSpi) super.clone();
		copy.buffer = copy.buffer.clone();
		return copy;
	}
}
