/*
 * Copyright (c) 2003, 2014, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2017-2022 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.is2t.support.security.mac;

import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.Key;
import java.security.spec.AlgorithmParameterSpec;

import javax.crypto.MacSpi;
import javax.crypto.SecretKey;

abstract public class AbstractMacSpi extends MacSpi implements Cloneable {

	// one element byte array, temporary storage for update(byte)
	private byte[] oneByte;

	// length of the message digest in bytes
	private final int macLength;

	private boolean isReset; // Is this the first data to be processed?

	/**
	 * Main constructor.
	 */
	AbstractMacSpi(int macLength) {
		super();
		this.macLength = macLength;
		this.isReset = true;
	}

	@Override
	protected final int engineGetMacLength() {
		return this.macLength;
	}

	@Override
	protected void engineInit(Key key, AlgorithmParameterSpec params)
			throws InvalidKeyException, InvalidAlgorithmParameterException {
		if (params != null) {
			throw new InvalidAlgorithmParameterException("HMAC does not use parameters");
		}

		if (!(key instanceof SecretKey)) {
			throw new InvalidKeyException("Secret key expected");
		}

		byte[] secret = key.getEncoded();
		if (secret == null) {
			throw new InvalidKeyException("Missing key data");
		}

		implInit(secret);
	}

	// single byte update. See JCA doc.
	@Override
	protected final void engineUpdate(byte b) {
		if (this.oneByte == null) {
			this.oneByte = new byte[1];
		}
		this.oneByte[0] = b;
		engineUpdate(this.oneByte, 0, 1);
	}

	// array update. See JCA doc.
	@Override
	protected final void engineUpdate(byte[] b, int ofs, int len) {
		if (len == 0) {
			return;
		}
		if ((ofs < 0) || (len < 0) || (ofs > b.length - len)) {
			throw new ArrayIndexOutOfBoundsException();
		}
		if (!this.isReset) {
			engineReset();
		}

		implUpdate(b, ofs, len);
	}

	@Override
	protected final byte[] engineDoFinal() {
		byte[] b = new byte[this.macLength];
		if (!this.isReset) {
			engineReset();
		}
		implDoFinal(b, 0, b.length);
		this.isReset = false;

		return b;
	}

	@Override
	protected void engineReset() {
		if (this.isReset) {
			// already reset, ignore
			return;
		}
		implReset();
		this.isReset = true;
	}

	abstract void implInit(byte[] secretKey) throws InvalidKeyException;

	/**
	 * TODO comment
	 */
	abstract void implUpdate(byte[] b, int ofs, int len);

	abstract void implDoFinal(byte[] out, int ofs, int len);

	/**
	 * Reset subclass specific state to their initial values. AbstractMacSpi calls this method when necessary.
	 */
	abstract void implReset();

}
