/*
 * Copyright (c) 2015, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2017-2023 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.is2t.support.security.x509;

import java.io.IOException;
import java.security.PublicKey;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateException;
import java.security.cert.CertificateExpiredException;
import java.security.cert.CertificateNotYetValidException;

import com.is2t.support.security.NativeKey;
import com.is2t.support.security.SecurityErrorConstants;
import com.is2t.support.security.natives.X509CertSupportNatives;
import com.is2t.support.security.x509.X509CertImpl.EncodedFormat;

import ej.sni.NativeException;

public class X509CertSupport {

	/**
	 * Checks that the specified date is within the certificate's validity period, or basically if the certificate would
	 * be valid at the specified date/time.
	 *
	 * @param cert
	 *            the certificate
	 * @throws CertificateEncodingException
	 *             if the certificate encoding is not recognized
	 * @exception CertificateExpiredException
	 *                if the certificate has expired.
	 * @throws CertificateNotYetValidException
	 *             if the certificate is not yet valid.
	 */
	public static void checkValidity(X509CertImpl cert)
			throws CertificateExpiredException, CertificateNotYetValidException {
		byte[] data;
		try {
			data = cert.getEncoded();
		} catch (CertificateEncodingException e) {
			throw new NativeException(SecurityErrorConstants.J_UNKNOWN_ERROR, e.getMessage());
		}
		final int validityCheckResult = X509CertSupportNatives.checkValidity(data, data.length);
		if (validityCheckResult == SecurityErrorConstants.J_SEC_NO_ERROR) {
			// all good - this method exits, we want to check exactly the returned value for "no error" and not
			// treat it on an else case
		} else if (validityCheckResult == SecurityErrorConstants.J_X509_CERT_EXPIRED_ERROR) {
			throw new CertificateExpiredException("The certificate expired.");
		} else if (validityCheckResult == SecurityErrorConstants.J_X509_CERT_NOT_YET_VALID_ERROR) {
			throw new CertificateNotYetValidException("The certificate is not yet valid.");
		} else {
			throw new IllegalStateException("Unexpected error, error code: " + validityCheckResult);
		}
	}

	/*
	 * Parse the X509 certificate and returns the certificate encoded format
	 */
	public static EncodedFormat parse(X509CertImpl cert) throws CertificateException, IOException {
		byte[] data = cert.getEncoded();
		int ret = X509CertSupportNatives.parse(data, 0, data.length);

		if (ret < 0) {
			throw new IOException("Unable to parse certificate");
		}
		switch (ret) {
		case 0:
			return EncodedFormat.PEM;
		case 1:
			return EncodedFormat.DER;
		default:
			throw new IOException("Unsupported certificate format. Only DER and PEM enconded format are supported.");
		}
	}

	public static void verify(X509CertImpl cert, PublicKey key) throws CertificateEncodingException {
		byte[] data = cert.getEncoded();
		int nativeId = ((NativeKey) key).getNativeId();
		X509CertSupportNatives.verify(data, data.length, nativeId);
	}
}
