/*
 * Copyright (c) 1997, 2023, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2024 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package com.microej.support.security.pbekey;

import java.security.MessageDigest;
import java.security.spec.InvalidKeySpecException;

import javax.crypto.SecretKey;
import javax.crypto.spec.PBEKeySpec;
import javax.security.auth.DestroyFailedException;

import com.microej.security.util.Arrays;

/**
 * Implementation copied from <code>com.sun.crypto.provider.PBEKey</code>.
 */

/**
 * This class represents a PBE key.
 *
 * @author Jan Luehe
 *
 */
final class PBEKey implements SecretKey {

	private static final long serialVersionUID = -2234768909660948176L;

	private final byte[] key;

	private final String type;

	private boolean destroyed;

	/**
	 * Creates a PBE key from a given PBE key specification.
	 *
	 * @param keytype
	 *            the given PBE key specification
	 */
	PBEKey(PBEKeySpec keySpec, String keytype) throws InvalidKeySpecException {
		char[] passwd = keySpec.getPassword();
		if (passwd == null) {
			// Should allow an empty password.
			passwd = new char[0];
		}
		// Accept "\0" to signify "zero-length password with no terminator".
		if (!(passwd.length == 1 && passwd[0] == 0)) {
			for (int i = 0; i < passwd.length; i++) {
				if ((passwd[i] < '\u0020') || (passwd[i] > '\u007E')) {
					throw new InvalidKeySpecException("Password is not ASCII");
				}
			}
		}
		this.key = new byte[passwd.length];
		for (int i = 0; i < passwd.length; i++) {
			this.key[i] = (byte) (passwd[i] & 0x7f);
		}
		Arrays.fill(passwd, '\0');
		this.type = keytype;
	}

	@Override
	public byte[] getEncoded() {
		return this.key.clone();
	}

	@Override
	public String getAlgorithm() {
		return this.type;
	}

	@Override
	public String getFormat() {
		return "RAW";
	}

	/**
	 * Calculates a hash code value for the object. Objects that are equal will also have the same hashcode.
	 */
	@Override
	public int hashCode() {
		return Arrays.hashCode(this.key) ^ getAlgorithm().toLowerCase().hashCode();
	}

	@Override
	public boolean equals(Object obj) {
		if (obj == this) {
			return true;
		}

		if (!(obj instanceof SecretKey)) {
			return false;
		}

		SecretKey that = (SecretKey) obj;

		// destroyed keys are considered different
		if (isDestroyed() || that.isDestroyed()) {
			return false;
		}

		if (!(that.getAlgorithm().equalsIgnoreCase(this.type))) {
			return false;
		}

		byte[] thatEncoded = that.getEncoded();
		boolean ret = MessageDigest.isEqual(this.key, thatEncoded);
		Arrays.fill(thatEncoded, (byte) 0x00);
		return ret;
	}

	@Override
	public void destroy() throws DestroyFailedException {
		if (!this.destroyed) {
			this.destroyed = true;
			Arrays.fill(this.key, (byte) 0x00);
		}
	}

	@Override
	public boolean isDestroyed() {
		return this.destroyed;
	}
}
