/*
 * Copyright (c) 1997, 2023, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2024 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package com.microej.support.security.pbekey;

import java.security.spec.InvalidKeySpecException;
import java.security.spec.KeySpec;
import java.util.ArrayList;

import javax.crypto.SecretKey;
import javax.crypto.SecretKeyFactorySpi;
import javax.crypto.spec.PBEKeySpec;

/**
 * Implementation copied from <code>com.sun.crypto.provider.PBEKeyFactory</code>.
 */

/**
 * This class implements a key factory for PBE keys according to PKCS#5, meaning that the password must consist of
 * printable ASCII characters (values 32 to 126 decimal inclusive) and only the low order 8 bits of each password
 * character are used.
 *
 * @author Jan Luehe
 *
 */
public class PBEKeyFactory extends SecretKeyFactorySpi {

	private final String type;
	private static final ArrayList<String> validTypes;

	/**
	 * Simple constructor.
	 *
	 * @param keytype
	 *            one of the supported key types.
	 */
	public PBEKeyFactory(String keytype) {
		this.type = keytype;
	}

	static {
		validTypes = new ArrayList<>();
		// Only PKCS12 PBE ciphers with 3DES cipher are supported.
		validTypes.add("PBEWithSHA1AndDESede".toUpperCase());
		// Only PBES2 ciphers without HMAC output truncation and with AES_128 cipher are supported.
		validTypes.add("PBEWithHmacSHA1AndAES_128".toUpperCase());
		validTypes.add("PBEWithHmacSHA224AndAES_128".toUpperCase());
		validTypes.add("PBEWithHmacSHA256AndAES_128".toUpperCase());
		validTypes.add("PBEWithHmacSHA384AndAES_128".toUpperCase());
		validTypes.add("PBEWithHmacSHA512AndAES_128".toUpperCase());
	}

	/**
	 * Generates a <code>SecretKey</code> object from the provided key specification (key material).
	 *
	 * @param keySpec
	 *            the specification (key material) of the secret key
	 *
	 * @return the secret key
	 *
	 * @exception InvalidKeySpecException
	 *                if the given key specification is inappropriate for this key factory to produce a public key.
	 */
	@Override
	protected SecretKey engineGenerateSecret(KeySpec keySpec) throws InvalidKeySpecException {
		if (!(keySpec instanceof PBEKeySpec)) {
			throw new InvalidKeySpecException("Invalid key spec");
		}
		return new PBEKey((PBEKeySpec) keySpec, this.type);
	}

	/**
	 * Checks if the algorithm is supported by the key factory.
	 *
	 * @param algorithm
	 *            string that describes the algorithm.
	 * @return true if the algorithm is supported by the key factory, false otherwise.
	 */
	public static boolean isAlgorithmSupported(String algorithm) {
		return ((algorithm != null) && (validTypes.contains(algorithm.toUpperCase())));
	}
}
