/*
 * Copyright (c) 1996, 2020, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2018-2022 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package sun.security.x509;

import java.io.Serializable;
import java.security.NoSuchAlgorithmException;

import com.is2t.support.security.signature.NativeSignatureSpi;

import ej.sni.NativeException;
import ej.sni.SNI;

/**
 *
 * Use this class to get OID of supported algorithms see {@link AlgorithmId#get(String)}
 *
 * The OID is handled in it's String format in this class. example 1.2.840.113549.1.1.11 for "SHA256withRSA"
 *
 * @author David Brownell
 * @author Amit Kapoor
 * @author Hemma Prafullchandra
 */
public class AlgorithmId implements Serializable {

	/** use serialVersionUID from JDK 1.1. for interoperability */
	private static final long serialVersionUID = 7205873507486557157L;

	/**
	 * The object identitifer being used for this algorithm.
	 */
	private String algid;

	/**
	 * Constructs an algorithm ID which will be initialized separately, for example by deserialization.
	 *
	 * @deprecated use one of the other constructors.
	 */
	@Deprecated
	public AlgorithmId() {
	}

	/**
	 * Constructs a parameterless algorithm ID.
	 *
	 * @param oid
	 *            the identifier for the algorithm
	 */
	public AlgorithmId(String oid) {
		this.algid = oid;
	}

	/**
	 * Returns the ISO OID for this algorithm. This is usually converted to a string and used as part of an algorithm
	 * name, for example "OID.1.3.14.3.2.13" style notation. Use the <code>getName</code> call when you do not need to
	 * ensure cross-system portability of algorithm names, or need a user friendly name.
	 */
	public final String getOID() {
		return this.algid;
	}

	/**
	 * Returns true iff the argument indicates the same algorithm with the same parameters.
	 */
	public boolean equals(AlgorithmId other) {
		return (this.algid.equals(other.algid));
	}

	/**
	 * Compares this AlgorithmID to another. If algorithm parameters are available, they are compared. Otherwise, just
	 * the object IDs for the algorithm are compared.
	 *
	 * @param other
	 *            preferably an AlgorithmId, else an ObjectIdentifier
	 */
	@Override
	public boolean equals(Object other) {
		if (this == other) {
			return true;
		}
		if (other instanceof AlgorithmId) {
			return equals((AlgorithmId) other);
		} else {
			return false;
		}
	}

	/**
	 * Compares two algorithm IDs for equality. Returns true iff they are the same algorithm, ignoring algorithm
	 * parameters.
	 */
	public final boolean equals(String id) {
		return this.algid.equals(id);
	}

	/**
	 * Returns a hashcode for this AlgorithmId.
	 *
	 * @return a hashcode for this AlgorithmId.
	 */
	@Override
	public int hashCode() {
		return this.algid.hashCode();
	}

	/**
	 * Returns one of the algorithm IDs most commonly associated with this algorithm name.
	 *
	 * @param algname
	 *            the name being used
	 * @deprecated use the short get form of this method.
	 * @exception NoSuchAlgorithmException
	 *                on error.
	 */
	@Deprecated
	public static AlgorithmId getAlgorithmId(String algname) throws NoSuchAlgorithmException {
		return get(algname);
	}

	/**
	 * Returns one of the algorithm IDs most commonly associated with this algorithm name.
	 *
	 * @param algname
	 *            the name being used
	 * @exception NoSuchAlgorithmException
	 *                on error.
	 */
	public static AlgorithmId get(String algname) throws NoSuchAlgorithmException {
		String oid;
		try {
			byte[] cAlgorithm = new byte[algname.length() + 1];// add 1 for the '\0'
			SNI.toCString(algname, cAlgorithm);

			byte[] output = new byte[256];// OID max length is to 128 we use x2

			NativeSignatureSpi.getAlgOID(cAlgorithm, output, output.length);

			oid = SNI.toJavaString(output);
		} catch (NativeException ioe) {
			throw new NoSuchAlgorithmException("Invalid ObjectIdentifier " + algname);
		}

		if (oid == null) {
			throw new NoSuchAlgorithmException("unrecognized algorithm name: " + algname);
		}
		return new AlgorithmId(oid);
	}

	/**
	 * Creates a signature algorithm name from a digest algorithm name and a encryption algorithm name.
	 */
	public static String makeSigAlg(String digAlg, String encAlg) {
		digAlg = digAlg.replace("-", "");
		if (encAlg.equalsIgnoreCase("EC")) {
			encAlg = "ECDSA";
		}

		return digAlg + "with" + encAlg;
	}

	/**
	 * Extracts the encryption algorithm name from a signature algorithm name.
	 */
	public static String getEncAlgFromSigAlg(String signatureAlgorithm) {
		signatureAlgorithm = signatureAlgorithm.toUpperCase();
		int with = signatureAlgorithm.indexOf("WITH");
		String keyAlgorithm = null;
		if (with > 0) {
			int and = signatureAlgorithm.indexOf("AND", with + 4);
			if (and > 0) {
				keyAlgorithm = signatureAlgorithm.substring(with + 4, and);
			} else {
				keyAlgorithm = signatureAlgorithm.substring(with + 4);
			}
			if (keyAlgorithm.equalsIgnoreCase("ECDSA")) {
				keyAlgorithm = "EC";
			}
		}
		return keyAlgorithm;
	}

	/**
	 * Extracts the digest algorithm name from a signature algorithm name.
	 */
	public static String getDigAlgFromSigAlg(String signatureAlgorithm) {
		signatureAlgorithm = signatureAlgorithm.toUpperCase();
		int with = signatureAlgorithm.indexOf("WITH");
		if (with > 0) {
			return signatureAlgorithm.substring(0, with);
		}
		return null;
	}
}
